import os
import unittest
import sqlite3
import datetime
import pandas

from gtfspy.gtfs import GTFS
from gtfspy.filter import FilterExtract
from gtfspy.filter import remove_all_trips_fully_outside_buffer

from gtfspy.import_gtfs import import_gtfs
import hashlib


class TestGTFSFilter(unittest.TestCase):

    def setUp(self):
        self.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        self.gtfs_source_dir_filter_test = os.path.join(self.gtfs_source_dir, "filter_test_feed/")

        # self.G = GTFS.from_directory_as_inmemory_db(self.gtfs_source_dir)

        # some preparations:
        self.fname = self.gtfs_source_dir + "/test_gtfs.sqlite"
        self.fname_copy = self.gtfs_source_dir + "/test_gtfs_copy.sqlite"
        self.fname_filter = self.gtfs_source_dir + "/test_gtfs_filter_test.sqlite"

        self._remove_temporary_files()
        self.assertFalse(os.path.exists(self.fname_copy))

        conn = sqlite3.connect(self.fname)
        import_gtfs(self.gtfs_source_dir, conn, preserve_connection=True, print_progress=False)
        conn_filter = sqlite3.connect(self.fname_filter)
        import_gtfs(self.gtfs_source_dir_filter_test, conn_filter, preserve_connection=True, print_progress=False)

        self.G = GTFS(conn)
        self.G_filter_test = GTFS(conn_filter)

        self.hash_orig = hashlib.md5(open(self.fname, 'rb').read()).hexdigest()

    def _remove_temporary_files(self):
        for fn in [self.fname, self.fname_copy, self.fname_filter]:
            if os.path.exists(fn) and os.path.isfile(fn):
                os.remove(fn)

    def tearDown(self):
        self._remove_temporary_files()

    def test_copy(self):
        # do a simple copy
        FilterExtract(self.G, self.fname_copy, update_metadata=False).create_filtered_copy()

        # check that the copying has been properly performed:
        hash_copy = hashlib.md5(open(self.fname_copy, 'rb').read()).hexdigest()
        self.assertTrue(os.path.exists(self.fname_copy))
        self.assertEqual(self.hash_orig, hash_copy)

    def test_filter_change_metadata(self):
        # A simple test that changing update_metadata to True, does update some stuff:
        FilterExtract(self.G, self.fname_copy, update_metadata=True).create_filtered_copy()
        # check that the copying has been properly performed:
        hash_orig = hashlib.md5(open(self.fname, 'rb').read()).hexdigest()
        hash_copy = hashlib.md5(open(self.fname_copy, 'rb').read()).hexdigest()
        self.assertTrue(os.path.exists(self.fname_copy))
        self.assertNotEqual(hash_orig, hash_copy)
        os.remove(self.fname_copy)

    def test_filter_by_agency(self):
        FilterExtract(self.G, self.fname_copy, agency_ids_to_preserve=['DTA']).create_filtered_copy()
        hash_copy = hashlib.md5(open(self.fname_copy, 'rb').read()).hexdigest()
        self.assertNotEqual(self.hash_orig, hash_copy)
        G_copy = GTFS(self.fname_copy)
        agency_table = G_copy.get_table("agencies")
        assert "EXA" not in agency_table['agency_id'].values, "EXA agency should not be preserved"
        assert "DTA" in agency_table['agency_id'].values, "DTA agency should be preserved"
        routes_table = G_copy.get_table("routes")
        assert "EXR1" not in routes_table['route_id'].values, "EXR1 route_id should not be preserved"
        assert "AB" in routes_table['route_id'].values, "AB route_id should be preserved"
        trips_table = G_copy.get_table("trips")
        assert "EXT1" not in trips_table['trip_id'].values, "EXR1 route_id should not be preserved"
        assert "AB1" in trips_table['trip_id'].values, "AB1 route_id should be preserved"
        calendar_table = G_copy.get_table("calendar")
        assert "FULLW" in calendar_table['service_id'].values, "FULLW service_id should be preserved"
        # stop_times
        stop_times_table = G_copy.get_table("stop_times")
        # 01:23:45 corresponds to 3600 + (32 * 60) + 45 [in day seconds]
        assert 3600 + (32 * 60) + 45 not in stop_times_table['arr_time']
        os.remove(self.fname_copy)


    def test_filter_by_start_and_end_full_range(self):
        # untested tables with filtering: stops, shapes
        # test filtering by start and end time, copy full range
        FilterExtract(self.G, self.fname_copy, start_date=u"2007-01-01", end_date=u"2011-01-01", update_metadata=False).create_filtered_copy()
        G_copy = GTFS(self.fname_copy)
        dsut_end = G_copy.get_day_start_ut("2010-12-31")
        dsut_to_trip_I = G_copy.get_tripIs_within_range_by_dsut(dsut_end, dsut_end + 24 * 3600)
        self.assertGreater(len(dsut_to_trip_I), 0)
        os.remove(self.fname_copy)

    def test_filter_end_date_not_included(self):
        # the end date should not be included:
        FilterExtract(self.G, self.fname_copy, start_date="2007-01-02", end_date="2010-12-31").create_filtered_copy()

        hash_copy = hashlib.md5(open(self.fname_copy, 'rb').read()).hexdigest()
        self.assertNotEqual(self.hash_orig, hash_copy)
        G_copy = GTFS(self.fname_copy)
        dsut_end = G_copy.get_day_start_ut("2010-12-31")
        dsut_to_trip_I = G_copy.get_tripIs_within_range_by_dsut(dsut_end, dsut_end + 24 * 3600)
        self.assertEqual(len(dsut_to_trip_I), 0)

        calendar_copy = G_copy.get_table("calendar")
        max_date_calendar = max([datetime.datetime.strptime(el, "%Y-%m-%d")
                                 for el in calendar_copy["end_date"].values])
        min_date_calendar = max([datetime.datetime.strptime(el, "%Y-%m-%d")
                                 for el in calendar_copy["start_date"].values])
        end_date_not_included = datetime.datetime.strptime("2010-12-31", "%Y-%m-%d")
        start_date_not_included = datetime.datetime.strptime("2007-01-01", "%Y-%m-%d")
        self.assertLess(max_date_calendar, end_date_not_included, msg="the last date should not be included in calendar")
        self.assertLess(start_date_not_included, min_date_calendar)
        os.remove(self.fname_copy)

    def test_filter_spatially(self):
        # test that the db is split by a given spatial boundary
        FilterExtract(self.G, self.fname_copy, buffer_lat=36.914893, buffer_lon=-116.76821, buffer_distance_km=50).create_filtered_copy()
        G_copy = GTFS(self.fname_copy)

        stops_table = G_copy.get_table("stops")
        self.assertNotIn("FUR_CREEK_RES", stops_table['stop_id'].values)
        self.assertIn("AMV", stops_table['stop_id'].values)
        self.assertEqual(len(stops_table['stop_id'].values), 8)

        conn_copy = sqlite3.connect(self.fname_copy)
        stop_ids_df = pandas.read_sql('SELECT stop_id from stop_times '
                                      'left join stops '
                                      'on stops.stop_I = stop_times.stop_I', conn_copy)
        stop_ids = stop_ids_df["stop_id"].values

        self.assertNotIn("FUR_CREEK_RES", stop_ids)
        self.assertIn("AMV", stop_ids)

        trips_table = G_copy.get_table("trips")
        self.assertNotIn("BFC1", trips_table['trip_id'].values)

        routes_table = G_copy.get_table("routes")
        self.assertNotIn("BFC", routes_table['route_id'].values)
        # cases:
        # whole trip excluded
        # whole route excluded
        # whole agency excluded
        # part of trip excluded
        # part of route excluded
        # part of agency excluded
        # not removing stops from a trip that returns into area

        # test higher-order removals
        # stop A preserved
        # -> stop B preserved
        # -> stop C preserved

    def test_filter_spatially_2(self):
        n_rows_before = {
            "routes": 4,
            "stop_times": 14,
            "trips": 4,
            "stops": 6,
            "shapes": 4
        }
        n_rows_after_1000 = { # within "soft buffer" in the feed data
            "routes": 1,
            "stop_times": 2,
            "trips": 1,
            "stops": 2,
            "shapes": 0
        }
        n_rows_after_3000 = { # within "hard buffer" in the feed data
            "routes": len(["t1", "t3", "t4"]),
            "stop_times": 11,
            "trips": 4,
            "stops": len({"P", "H", "V", "L", "B"}),
            # for some reason, the first "shapes": 4
        }
        paris_lat = 48.832781
        paris_lon = 2.360734

        SELECT_MIN_MAX_SHAPE_BREAKS_BY_TRIP_I_SQL = \
            "SELECT trips.trip_I, shape_id, min(shape_break) as min_shape_break, max(shape_break) as max_shape_break FROM trips, stop_times WHERE trips.trip_I=stop_times.trip_I GROUP BY trips.trip_I"
        trip_min_max_shape_seqs = pandas.read_sql(SELECT_MIN_MAX_SHAPE_BREAKS_BY_TRIP_I_SQL, self.G_filter_test.conn)

        for distance_km, n_rows_after in zip([1000, 3000], [n_rows_after_1000, n_rows_after_3000]):
            try:
                os.remove(self.fname_copy)
            except FileNotFoundError:
                pass
            FilterExtract(self.G_filter_test,
                          self.fname_copy,
                          buffer_lat=paris_lat,
                          buffer_lon=paris_lon,
                          buffer_distance_km=distance_km).create_filtered_copy()
            for table_name, n_rows in n_rows_before.items():
                self.assertEqual(len(self.G_filter_test.get_table(table_name)), n_rows, "Row counts before differ in " + table_name + ", distance: " + str(distance_km))
            G_copy = GTFS(self.fname_copy)
            for table_name, n_rows in n_rows_after.items():
                table = G_copy.get_table(table_name)
                self.assertEqual(len(table), n_rows, "Row counts after differ in " + table_name + ", distance: " + str(distance_km) + "\n" + str(table))

            # assert that stop_times are resequenced starting from one
            counts = pandas.read_sql("SELECT count(*) FROM stop_times GROUP BY trip_I ORDER BY trip_I", G_copy.conn)
            max_values = pandas.read_sql("SELECT max(seq) FROM stop_times GROUP BY trip_I ORDER BY trip_I", G_copy.conn)
            self.assertTrue((counts.values == max_values.values).all())

    def test_remove_all_trips_fully_outside_buffer(self):
        stops = self.G.stops()
        stop_1 = stops[stops['stop_I'] == 1]

        n_trips_before = len(self.G.get_table("trips"))

        remove_all_trips_fully_outside_buffer(self.G.conn, float(stop_1.lat), float(stop_1.lon), 100000)
        self.assertEqual(len(self.G.get_table("trips")), n_trips_before)

        # 0.002 (=max 2 meters from the stop), rounding errors can take place...
        remove_all_trips_fully_outside_buffer(self.G.conn, float(stop_1.lat), float(stop_1.lon), 0.002)
        self.assertEqual(len(self.G.get_table("trips")), 2)  # value "2" comes from the data






import os

import networkx
import pandas
from osmread import parse_file, Way, Node

from gtfspy.gtfs import GTFS
from gtfspy.util import wgs84_distance

from warnings import warn

from geoindex import GeoGridIndex, GeoPoint


def add_walk_distances_to_db_python(gtfs, osm_path, cutoff_distance_m=1000):
    """
    Computes the walk paths between stops, and updates these to the gtfs database.

    Parameters
    ----------
    gtfs: gtfspy.GTFS or str
        A GTFS object or a string representation.
    osm_path: str
        path to the OpenStreetMap file
    cutoff_distance_m: number
        maximum allowed distance in meters

    Returns
    -------
    None

    See Also
    --------
    gtfspy.calc_transfers
    compute_walk_paths_java
    """
    if isinstance(gtfs, str):
        gtfs = GTFS(gtfs)
    assert (isinstance(gtfs, GTFS))
    print("Reading in walk network")
    walk_network = create_walk_network_from_osm(osm_path)
    print("Matching stops to the OSM network")
    stop_I_to_nearest_osm_node, stop_I_to_nearest_osm_node_distance = match_stops_to_nodes(gtfs, walk_network)

    transfers = gtfs.get_straight_line_transfer_distances()

    from_I_to_to_stop_Is = {stop_I: set() for stop_I in stop_I_to_nearest_osm_node}
    for transfer_tuple in transfers.itertuples():
        from_I = transfer_tuple.from_stop_I
        to_I = transfer_tuple.to_stop_I
        from_I_to_to_stop_Is[from_I].add(to_I)

    print("Computing walking distances")
    for from_I, to_stop_Is in from_I_to_to_stop_Is.items():
        from_node = stop_I_to_nearest_osm_node[from_I]
        from_dist = stop_I_to_nearest_osm_node_distance[from_I]
        shortest_paths = networkx.single_source_dijkstra_path_length(walk_network,
                                                                     from_node,
                                                                     cutoff=cutoff_distance_m - from_dist,
                                                                     weight="distance")
        for to_I in to_stop_Is:
            to_distance = stop_I_to_nearest_osm_node_distance[to_I]
            to_node = stop_I_to_nearest_osm_node[to_I]
            osm_distance = shortest_paths.get(to_node, float('inf'))
            total_distance = from_dist + osm_distance + to_distance
            from_stop_I_transfers = transfers[transfers['from_stop_I'] == from_I]
            straigth_distance = from_stop_I_transfers[from_stop_I_transfers["to_stop_I"] == to_I]["d"].values[0]
            assert (straigth_distance < total_distance + 2)  # allow for a maximum  of 2 meters in calculations
            if total_distance <= cutoff_distance_m:
                gtfs.conn.execute("UPDATE stop_distances "
                                  "SET d_walk = " + str(int(total_distance)) +
                                  " WHERE from_stop_I=" + str(from_I) + " AND to_stop_I=" + str(to_I))

    gtfs.conn.commit()


def match_stops_to_nodes(gtfs, walk_network):
    """
    Parameters
    ----------
    gtfs : a GTFS object
    walk_network : networkx.Graph

    Returns
    -------
    stop_I_to_node: dict
        maps stop_I to closest walk_network node
    stop_I_to_dist: dict
        maps stop_I to the distance to the closest walk_network node
    """
    network_nodes = walk_network.nodes(data="true")

    stop_Is = set(gtfs.get_straight_line_transfer_distances()['from_stop_I'])
    stops_df = gtfs.stops()

    geo_index = GeoGridIndex(precision=6)
    for net_node, data in network_nodes:
        geo_index.add_point(GeoPoint(data['lat'], data['lon'], ref=net_node))
    stop_I_to_node = {}
    stop_I_to_dist = {}
    for stop_I in stop_Is:
        stop_lat = float(stops_df[stops_df.stop_I == stop_I].lat)
        stop_lon = float(stops_df[stops_df.stop_I == stop_I].lon)
        geo_point = GeoPoint(stop_lat, stop_lon)
        min_dist = float('inf')
        min_dist_node = None
        search_distances_m = [0.100, 0.500]
        for search_distance_m in search_distances_m:
            for point, distance in geo_index.get_nearest_points(geo_point, search_distance_m, "km"):
                if distance < min_dist:
                    min_dist = distance * 1000
                    min_dist_node = point.ref
            if min_dist_node is not None:
                break
        if min_dist_node is None:
            warn("No OSM node found for stop: " + str(stops_df[stops_df.stop_I == stop_I]))
        stop_I_to_node[stop_I] = min_dist_node
        stop_I_to_dist[stop_I] = min_dist
    return stop_I_to_node, stop_I_to_dist


OSM_HIGHWAY_WALK_TAGS = {"trunk", "trunk_link", "primary", "primary_link", "secondary", "secondary_link", "tertiary",
                         "tertiary_link", "unclassified", "residential", "living_street", "road", "pedestrian", "path",
                         "cycleway", "footway"}


def create_walk_network_from_osm(osm_file):
    walk_network = networkx.Graph()
    assert (os.path.exists(osm_file))
    ways = []
    for i, entity in enumerate(parse_file(osm_file)):
        if isinstance(entity, Node):
            walk_network.add_node(entity.id, lat=entity.lat, lon=entity.lon)
        elif isinstance(entity, Way):
            if "highway" in entity.tags:
                if entity.tags["highway"] in OSM_HIGHWAY_WALK_TAGS:
                    ways.append(entity)
    for way in ways:
        walk_network.add_path(way.nodes)
    del ways

    # Remove all singleton nodes (note that taking the giant component does not necessarily provide proper results.
    for node, degree in walk_network.degree().items():
        if degree is 0:
            walk_network.remove_node(node)

    node_lats = networkx.get_node_attributes(walk_network, 'lat')
    node_lons = networkx.get_node_attributes(walk_network, 'lon')
    for source, dest, data in walk_network.edges(data=True):
        data["distance"] = wgs84_distance(node_lats[source],
                                          node_lons[source],
                                          node_lats[dest],
                                          node_lons[dest])
    return walk_network


def compute_walk_paths_java(gtfs_db_path, osm_file, cache_db=None):
    """
    Parameters
    ----------
    gtfs_db_path: str (path to the gtfs database)
    osm_file: str
    cache_db: str

    Returns
    -------
    None
    """
    raise NotImplementedError("This has not been pipelined yet. Please see the Python code.")



import pandas

from gtfspy import util
from gtfspy.import_loaders.table_loader import TableLoader, decode_six


class FrequenciesLoader(TableLoader):
    """Load the general frequency table."""
    fname = 'frequencies.txt'
    table = 'frequencies'

    tabledef = (u'(trip_I INT, '
                u'start_time TEXT, '
                u'end_time TEXT, '
                u'headway_secs INT,'
                u'exact_times INT, '
                u'start_time_ds INT, '
                u'end_time_ds INT'
                u')')
    extra_keys = [u'trip_I',
                  u'start_time_ds',
                  u'end_time_ds',
                  ]
    extra_values = [u'(SELECT trip_I FROM trips WHERE trip_id=:_trip_id )',
                    '(substr(:start_time,-8,2)*3600 + substr(:start_time,-5,2)*60 + substr(:start_time,-2))',
                    '(substr(:end_time,-8,2)*3600 + substr(:end_time,-5,2)*60 + substr(:end_time,-2))',
                    ]

    def gen_rows(self, readers, prefixes):
        for reader, prefix in zip(readers, prefixes):
            for row in reader:
                yield dict(
                    _trip_id=prefix + decode_six(row['trip_id']),
                    start_time=row['start_time'],
                    end_time=row['end_time'],
                    headway_secs=int(row['headway_secs']),
                    exact_times=int(row['exact_times']) if 'exact_times' in row and row['exact_times'].isdigit() else 0
                )

    def post_import(self, cur):
        # For each (start_time_dependent) trip_I in frequencies.txt
        conn = self._conn
        frequencies_df = pandas.read_sql("SELECT * FROM " + self.table, conn)

        for freq_tuple in frequencies_df.itertuples():
            trip_data = pandas.read_sql_query("SELECT * FROM trips WHERE trip_I= " + str(int(freq_tuple.trip_I)), conn)
            assert len(trip_data) == 1
            trip_data = list(trip_data.itertuples())[0]
            freq_start_time_ds = freq_tuple.start_time_ds
            freq_end_time_ds = freq_tuple.end_time_ds
            trip_duration = cur.execute("SELECT max(arr_time_ds) - min(dep_time_ds) "
                                        "FROM stop_times "
                                        "WHERE trip_I={trip_I}".format(trip_I=str(int(freq_tuple.trip_I)))
                                        ).fetchone()[0]
            if trip_duration is None:
                raise ValueError("Stop times for frequency trip " + trip_data.trip_id + " are not properly defined")
            headway = freq_tuple.headway_secs

            sql = "SELECT * FROM stop_times WHERE trip_I=" + str(trip_data.trip_I) + " ORDER BY seq"
            stop_time_data = pandas.read_sql_query(sql, conn)

            start_times_ds = range(freq_start_time_ds, freq_end_time_ds, headway)
            for i, start_time in enumerate(start_times_ds):
                trip_id = trip_data.trip_id + u"_freq_" + str(start_time)
                route_I = trip_data.route_I
                service_I = trip_data.service_I

                shape_id = trip_data.shape_id
                direction_id = trip_data.direction_id
                headsign = trip_data.headsign
                end_time_ds = start_time + trip_duration

                # insert these into trips
                query = "INSERT INTO trips (trip_id, route_I, service_I, shape_id, direction_id, " \
                            "headsign, start_time_ds, end_time_ds)" \
                        " VALUES (?, ?, ?, ?, ?, ?, ?, ?)"

                params = [trip_id, int(route_I), int(service_I), shape_id, direction_id, headsign, int(start_time), int(end_time_ds)]
                cur.execute(query, params)

                query = "SELECT trip_I FROM trips WHERE trip_id='{trip_id}'".format(trip_id=trip_id)
                trip_I = cur.execute(query).fetchone()[0]

                # insert into stop_times
                # TODO! get the original data
                dep_times_ds = stop_time_data['dep_time_ds']
                dep_times_ds = dep_times_ds - min(dep_times_ds) + start_time
                arr_times_ds = stop_time_data['arr_time_ds']
                arr_times_ds = arr_times_ds - min(arr_times_ds) + start_time
                shape_breaks_series = stop_time_data['shape_break']
                stop_Is = stop_time_data['stop_I']

                shape_breaks = []
                for shape_break in shape_breaks_series:
                    value = None
                    try:
                        value = int(shape_break)
                    except Exception:
                        pass
                    shape_breaks.append(value)

                for seq, (dep_time_ds, arr_time_ds, shape_break, stop_I) in enumerate(zip(dep_times_ds,
                                                                                          arr_times_ds,
                                                                                          shape_breaks,
                                                                                          stop_Is)):
                    arr_time_hour = int(arr_time_ds // 3600)
                    query = "INSERT INTO stop_times (trip_I, stop_I, arr_time, " \
                            "dep_time, seq, arr_time_hour, shape_break, arr_time_ds, dep_time_ds) " \
                            "VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)"
                    arr_time = util.day_seconds_to_str_time(arr_time_ds)
                    dep_time = util.day_seconds_to_str_time(dep_time_ds)

                    cur.execute(query, (int(trip_I), int(stop_I), arr_time, dep_time, int(seq + 1),
                                        int(arr_time_hour), shape_break, int(arr_time_ds), int(dep_time_ds)))

        trip_Is = frequencies_df['trip_I'].unique()
        for trip_I in trip_Is:
            for table in ["trips", "stop_times"]:
                cur.execute("DELETE FROM {table} WHERE trip_I={trip_I}".format(table=table, trip_I=trip_I))
        self._conn.commit()


from heapq import heappush, heappop

import numpy as np

from gtfspy.route_types import WALK
from .event import Event

class EventHeap:
    """
    EventHeap represents a container for the event
    heap to run time-dependent Dijkstra for public transport routing objects.
    """

    def __init__(self, pd_df=None):
        """
        Parameters
        ----------
        pd_df : Pandas.Dataframe
            Initial list of
        """
        self.heap = []
        keys = ['arr_time_ut', 'dep_time_ut', 'from_stop_I', 'to_stop_I', 'trip_I']

        # pd_df.iterrows() is slow as it creates new Series objects!
        n = len(pd_df)
        key_to_j = {}
        for j, key in enumerate(pd_df.columns.values):
            key_to_j[key] = j
        pd_df_values = pd_df.values
        for i in range(n):
            vals = []
            for key in keys:
                j = key_to_j[key]
                vals.append(pd_df_values[i, j])
            e = Event(*vals)
            self.add_event(e)

    def add_event(self, event):
        """
        Add an event to the heap/priority queue

        Parameters
        ----------
        event : Event
        """
        assert event.dep_time_ut <= event.arr_time_ut
        heappush(self.heap, event)

    def pop_next_event(self):
        return heappop(self.heap)

    def size(self):
        """
        Return the size of the heap
        """
        return len(self.heap)

    def add_walk_events_to_heap(self, transfer_distances, e, start_time_ut, walk_speed, uninfected_stops, max_duration_ut):
        """
        Parameters
        ----------
        transfer_distances:
        e : Event
        start_time_ut : int
        walk_speed : float
        uninfected_stops : list
        max_duration_ut : int
        """
        n = len(transfer_distances)
        dists_values = transfer_distances.values
        to_stop_I_index = np.nonzero(transfer_distances.columns == 'to_stop_I')[0][0]
        d_index = np.nonzero(transfer_distances.columns == 'd')[0][0]
        for i in range(n):
            transfer_to_stop_I = dists_values[i, to_stop_I_index]
            if transfer_to_stop_I in uninfected_stops:
                d = dists_values[i, d_index]
                transfer_arr_time = e.arr_time_ut + int(d/float(walk_speed))
                if transfer_arr_time > start_time_ut+max_duration_ut:
                    continue
                te = Event(transfer_arr_time, e.arr_time_ut, e.to_stop_I, transfer_to_stop_I, WALK)
                self.add_event(te)



import matplotlib.colors
import matplotlib.cm
import matplotlib.colorbar
import matplotlib.pyplot
import numpy

# colormaps: "viridis", "plasma_r","seismic"


def get_colormap(observable_name=None):
    if "diff_minutes" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-30, vmax=30)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "diff_number" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-4, vmax=4)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "diff_relative" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-0.7, vmax=0.7)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "diff_multiples" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-2, vmax=2)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "diff_simpson" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-0.5, vmax=0.5)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "diff_n_trips" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-10, vmax=10)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "diff_n_routes" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-5, vmax=5)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "simpson" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=0, vmax=1.0)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "n_trips" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=0, vmax=30)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "n_routes" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=0, vmax=15)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "delay_minutes" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-300, vmax=300)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "delay_seconds" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=-1500, vmax=1500)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    elif "n_locations" in observable_name:
        norm = matplotlib.colors.Normalize(vmin=0, vmax=2000)
        cmap = matplotlib.cm.get_cmap(name="viridis", lut=None)
    else:
        norm = matplotlib.colors.Normalize(vmin=-30, vmax=30)
        cmap = matplotlib.cm.get_cmap(name="seismic", lut=None)
    return cmap, norm


def get_colormap_with_params(vmin, vmax, name="seismic"):
    norm = matplotlib.colors.Normalize(vmin=vmin, vmax=vmax)
    cmap = matplotlib.cm.get_cmap(name=name, lut=None)
    return cmap, norm


def get_list_of_colors(values, observable_name=None):
    cmap, norm = get_colormap(observable_name)
    scalarmap = matplotlib.cm.ScalarMappable(norm=norm, cmap=cmap)
    colorvalues = []
    for value in values:
        colorvalue = scalarmap.to_rgba(value)
        colorvalues.append(colorvalue)
    return colorvalues, norm, cmap

def createcolorbar(cmap, norm):
    """Create a colourbar with limits of lwr and upr"""
    cax, kw = matplotlib.colorbar.make_axes(matplotlib.pyplot.gca())
    c = matplotlib.colorbar.ColorbarBase(cax, cmap=cmap, norm=norm)
    return c






import unittest
from math import isnan
from unittest import TestCase

from gtfspy.routing.label import LabelTimeWithBoardingsCount
from gtfspy.routing.node_profile_multiobjective import NodeProfileMultiObjective
from gtfspy.routing.node_profile_analyzer_time_and_veh_legs import NodeProfileAnalyzerTimeAndVehLegs


class TestNodeProfileAnalyzerTimeAndVehLegs(TestCase):

    def setUp(self):
        self.label_class = LabelTimeWithBoardingsCount

    def _get_analyzer(self, labels, start_time, end_time, walk_to_target_duration=float('inf')):
        dep_times = list(set(map(lambda el: el.departure_time, labels)))
        p = NodeProfileMultiObjective(dep_times=dep_times,
                                      walk_to_target_duration=walk_to_target_duration,
                                      label_class=LabelTimeWithBoardingsCount)
        for label in labels:
            p.update([label])
        p.finalize()
        analyzer = NodeProfileAnalyzerTimeAndVehLegs.from_profile(p, start_time, end_time)
        return analyzer

    def test_trip_duration_statistics_empty_profile(self):
        analyzer = self._get_analyzer([], 0, 10)

        self.assertTrue(isnan(analyzer.max_trip_n_boardings()))
        self.assertTrue(isnan(analyzer.min_trip_n_boardings()))
        self.assertTrue(isnan(analyzer.mean_trip_n_boardings()))
        self.assertTrue(isnan(analyzer.median_trip_n_boardings()))

    def test_temporal_distances_by_n_vehicles(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=12, n_boardings=4, first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=15, n_boardings=2, first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=17, n_boardings=1, first_leg_is_walk=False)
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=10)
        median_temporal_distances = analyzer.median_temporal_distances()
        self.assertEqual(len(median_temporal_distances), 4 + 1)
        for i in range(len(median_temporal_distances) - 1):
            assert(median_temporal_distances[i] >= median_temporal_distances[i + 1])

    def test_n_boardings_on_shortest_paths(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=12, n_boardings=4,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=10, n_boardings=2,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=12, n_boardings=0,
                                        first_leg_is_walk=False)
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=10)
        self.assertEqual(analyzer.mean_n_boardings_on_shortest_paths(), 3)
        self.assertEqual(analyzer.min_n_boardings_on_shortest_paths(), 2)
        self.assertEqual(analyzer.max_n_boardings_on_shortest_paths(), 4)

    def test_min_n_boardings(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=12, n_boardings=4,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=10, n_boardings=2,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=12, n_boardings=1,
                                        first_leg_is_walk=False)
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=10)
        self.assertEqual(analyzer.min_n_boardings(), 0)
        analyzer2 = self._get_analyzer(labels, 0, 10, walk_to_target_duration=float('inf'))
        self.assertEqual(analyzer2.min_n_boardings(), 1)

    def test_min_n_boardings_after_departure_time_2(self):
        # (This tests the bug experienced with the Jollas region)
        labels = [
            LabelTimeWithBoardingsCount(departure_time=12, arrival_time_target=14, n_boardings=2,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=11, arrival_time_target=12, n_boardings=3,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=10, n_boardings=4,
                                        first_leg_is_walk=False),
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=float('inf'))
        self.assertEqual(analyzer.min_n_boardings(), 2)



    def test_min_n_boardings_on_fastest_paths(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=12, n_boardings=4,
                                        first_leg_is_walk=False),
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=10)
        self.assertEqual(analyzer.min_n_boardings_on_shortest_paths(), 0)

    def test_mean_n_boardings_on_fastest_paths(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=6, n_boardings=1,
                                        first_leg_is_walk=False),
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=10)
        self.assertEqual(analyzer.mean_n_boardings_on_shortest_paths(), 0.5)

        labels = [
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=15,
                                        n_boardings=1, first_leg_is_walk=False),
        ]
        analyzer = self._get_analyzer(labels, 0, 10, walk_to_target_duration=10)
        self.assertEqual(analyzer.mean_n_boardings_on_shortest_paths(), 0.5)

    def test_mean_temporal_distance_with_min_n_boardings(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=22, n_boardings=4,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=24, n_boardings=2,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=26, n_boardings=1,
                                        first_leg_is_walk=False)
        ]
        analyzer = self._get_analyzer(labels, 0, 5, walk_to_target_duration=float('inf'))
        self.assertEqual(analyzer.mean_temporal_distance_with_min_n_boardings(), 2.5 + 5 + 26 - 10)

    def test_n_boardings_on_fastest_trip(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=22, n_boardings=4,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=24, n_boardings=2,
                                        first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=26, n_boardings=1,
                                        first_leg_is_walk=False)
        ]
        analyzer = self._get_analyzer(labels, start_time=0, end_time=11, walk_to_target_duration=float('inf'))
        self.assertEqual(analyzer.n_boardings_on_fastest_trip(), 4)

    @unittest.skip
    def test_plot(self):
        labels = [
            LabelTimeWithBoardingsCount(departure_time=20, arrival_time_target=22, n_boardings=5, first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=15, arrival_time_target=20, n_boardings=6, first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=14, arrival_time_target=21, n_boardings=5, first_leg_is_walk=False),
            LabelTimeWithBoardingsCount(departure_time=13, arrival_time_target=22, n_boardings=4, first_leg_is_walk=False),
            # LabelTimeWithBoardingsCount(departure_time=12, arrival_time_target=23, n_vehicle_legs=3),
            LabelTimeWithBoardingsCount(departure_time=11, arrival_time_target=24, n_boardings=2, first_leg_is_walk=True),
            LabelTimeWithBoardingsCount(departure_time=10, arrival_time_target=25, n_boardings=1, first_leg_is_walk=True),
            LabelTimeWithBoardingsCount(departure_time=5, arrival_time_target=10, n_boardings=1, first_leg_is_walk=True)
        ]
        analyzer = self._get_analyzer(labels, 0, 20, 35)
        print(fig)
        import matplotlib.pyplot as plt
        plt.show()



from unittest import TestCase

import pyximport

from gtfspy.routing.journey_data import JourneyDataManager
from gtfspy.routing.label import LabelTimeWithBoardingsCount

pyximport.install()
import shutil
import os
from gtfspy.import_gtfs import import_gtfs
from gtfspy.routing.travel_impedance_data_store import TravelImpedanceDataStore


class TestJourneyData(TestCase):
    # noinspection PyAttributeOutsideInit

    def _import_sample_gtfs_db(self):
        import_gtfs([os.path.join(os.path.dirname(__file__), "../../test/test_data/test_gtfs.zip")], self.gtfs_path)

    def _remove_routing_test_data_directory_if_exists(self):
        try:
            shutil.rmtree(self.routing_tmp_test_data_dir)
        except FileNotFoundError:
            pass

    def _create_routing_test_data_directory(self):
        if not os.path.exists(self.routing_tmp_test_data_dir):
            os.makedirs(self.routing_tmp_test_data_dir)

    def setUp(self):
        self.routing_tmp_test_data_dir = "./tmp_routing_test_data/"
        self.gtfs_path = os.path.join(self.routing_tmp_test_data_dir, "test_gtfs.sqlite")
        self.data_store_path = os.path.join(self.routing_tmp_test_data_dir, "test_data_store.sqlite")
        self._remove_routing_test_data_directory_if_exists()
        self._create_routing_test_data_directory()

        self._import_sample_gtfs_db()
        self.jdm = JourneyDataManager(self.gtfs_path,
                                      os.path.join(self.routing_tmp_test_data_dir, "test_journeys.sqlite"),
                                      routing_params={"track_vehicle_legs": True})

    def tearDown(self):
        self._remove_routing_test_data_directory_if_exists()

    def test_boardings_computations_based_on_journeys(self):
        # input some journeys
        destination_stop = 1
        origin_stop = 2
        self.jdm.import_journey_data_for_target_stop(destination_stop,
                                                     {origin_stop:
                                                        [LabelTimeWithBoardingsCount(1, 2, 1, True),
                                                         LabelTimeWithBoardingsCount(2, 3, 2, True)]},
                                                     enforce_synchronous_writes=True
                                                     )
        self.jdm.compute_and_store_travel_impedance_measures(0, 2, self.data_store_path)
        store = TravelImpedanceDataStore(self.data_store_path)
        df = store.read_data_as_dataframe("temporal_distance")
        self.assertAlmostEqual(df.iloc[0]["min"], 1)
        self.assertAlmostEqual(df.iloc[0]["mean"], 1.5)
        self.assertAlmostEqual(df.iloc[0]["max"], 2.0)
        self.assertIn(df.iloc[0]["median"],[1, 2, 1.0, 1.5, 2.0])


from __future__ import unicode_literals

import csv
import pandas as pd

import numpy
import sys
import os

from gtfspy.gtfs import GTFS
from gtfspy.util import wgs84_distance


def get_spatial_bounds(gtfs, as_dict=False):
    """
    Parameters
    ----------
    gtfs

    Returns
    -------
    min_lon: float
    max_lon: float
    min_lat: float
    max_lat: float
    """
    stats = get_stats(gtfs)
    lon_min = stats['lon_min']
    lon_max = stats['lon_max']
    lat_min = stats['lat_min']
    lat_max = stats['lat_max']
    if as_dict:
        return {'lon_min': lon_min, 'lon_max': lon_max, 'lat_min': lat_min, 'lat_max': lat_max}
    else:
        return lon_min, lon_max, lat_min, lat_max


def get_percentile_stop_bounds(gtfs, percentile):
    stops = gtfs.get_table("stops")
    percentile = min(percentile, 100 - percentile)
    lat_min = numpy.percentile(stops['lat'].values, percentile)
    lat_max = numpy.percentile(stops['lat'].values, 100 - percentile)
    lon_min = numpy.percentile(stops['lon'].values, percentile)
    lon_max = numpy.percentile(stops['lon'].values, 100 - percentile)
    return lon_min, lon_max, lat_min, lat_max


def get_median_lat_lon_of_stops(gtfs):
    """
    Get median latitude AND longitude of stops

    Parameters
    ----------
    gtfs: GTFS

    Returns
    -------
    median_lat : float
    median_lon : float
    """
    stops = gtfs.get_table("stops")
    median_lat = numpy.percentile(stops['lat'].values, 50)
    median_lon = numpy.percentile(stops['lon'].values, 50)
    return median_lat, median_lon

def get_centroid_of_stops(gtfs):
    """
    Get mean latitude AND longitude of stops

    Parameters
    ----------
    gtfs: GTFS

    Returns
    -------
    mean_lat : float
    mean_lon : float
    """
    stops = gtfs.get_table("stops")
    mean_lat = numpy.mean(stops['lat'].values)
    mean_lon = numpy.mean(stops['lon'].values)
    return mean_lat, mean_lon


def write_stats_as_csv(gtfs, path_to_csv, re_write=False):
    """
    Writes data from get_stats to csv file

    Parameters
    ----------
    gtfs: GTFS
    path_to_csv: str
        filepath to the csv file to be generated
    re_write:
        insted of appending, create a new one.
    """
    stats_dict = get_stats(gtfs)
    # check if file exist
    if re_write:
            os.remove(path_to_csv)
    
    #if not os.path.isfile(path_to_csv):
     #   is_new = True
    #else:
     #   is_new = False
    
    is_new = True
    mode = 'r' if os.path.exists(path_to_csv) else 'w+'
    with open(path_to_csv, mode) as csvfile:
        for line in csvfile:
           if line:
               is_new = False
           else:
               is_new = True

    with open(path_to_csv, 'a') as csvfile:
        if (sys.version_info > (3, 0)):
            delimiter = u","
        else:
            delimiter = b","
        statswriter = csv.writer(csvfile, delimiter=delimiter)
        # write column names if
        if is_new:
            statswriter.writerow([key for key in sorted(stats_dict.keys())])

        row_to_write = []
        # write stats row sorted by column name
        for key in sorted(stats_dict.keys()):
            row_to_write.append(stats_dict[key])
        statswriter.writerow(row_to_write)


def get_stats(gtfs):
    """
    Get basic statistics of the GTFS data.

    Parameters
    ----------
    gtfs: GTFS

    Returns
    -------
    stats: dict
        A dictionary of various statistics.
        Keys should be strings, values should be inputtable to a database (int, date, str, ...)
        (but not a list)
    """
    stats = {}
    # Basic table counts
    for table in ['agencies', 'routes', 'stops', 'stop_times', 'trips', 'calendar', 'shapes', 'calendar_dates',
                  'days', 'stop_distances', 'frequencies', 'feed_info', 'transfers']:
        stats["n_" + table] = gtfs.get_row_count(table)

    # Agency names
    agencies = gtfs.get_table("agencies")
    stats["agencies"] = "_".join(agencies['name'].values)

    # Stop lat/lon range
    stops = gtfs.get_table("stops")
    lats = stops['lat'].values
    lons = stops['lon'].values
    percentiles = [0, 10, 50, 90, 100]

    try:
        lat_percentiles = numpy.percentile(lats, percentiles)
    except IndexError:
        lat_percentiles = [None] * 5
    lat_min, lat_10, lat_median, lat_90, lat_max = lat_percentiles
    stats["lat_min"] = lat_min
    stats["lat_10"] = lat_10
    stats["lat_median"] = lat_median
    stats["lat_90"] = lat_90
    stats["lat_max"] = lat_max

    try:
        lon_percentiles = numpy.percentile(lons, percentiles)
    except IndexError:
        lon_percentiles = [None] * 5
    lon_min, lon_10, lon_median, lon_90, lon_max = lon_percentiles
    stats["lon_min"] = lon_min
    stats["lon_10"] = lon_10
    stats["lon_median"] = lon_median
    stats["lon_90"] = lon_90
    stats["lon_max"] = lon_max

    if len(lats) > 0:
        stats["height_km"] = wgs84_distance(lat_min, lon_median, lat_max, lon_median) / 1000.
        stats["width_km"] = wgs84_distance(lon_min, lat_median, lon_max, lat_median) / 1000.
    else:
        stats["height_km"] = None
        stats["width_km"] = None

    first_day_start_ut, last_day_start_ut = gtfs.get_day_start_ut_span()
    stats["start_time_ut"] = first_day_start_ut
    if last_day_start_ut is None:
        stats["end_time_ut"] = None
    else:
        # 28 (instead of 24) comes from the GTFS stANDard
        stats["end_time_ut"] = last_day_start_ut + 28 * 3600

    stats["start_date"] = gtfs.get_min_date()
    stats["end_date"] = gtfs.get_max_date()

    # Maximum activity day
    max_activity_date = gtfs.execute_custom_query(
        'SELECT count(*), date '
        'FROM days '
        'GROUP BY date '
        'ORDER BY count(*) DESC, date '
        'LIMIT 1;').fetchone()
    if max_activity_date:
        stats["max_activity_date"] = max_activity_date[1]
        max_activity_hour = gtfs.get_cursor().execute(
            'SELECT count(*), arr_time_hour FROM day_stop_times '
            'WHERE date=? GROUP BY arr_time_hour '
            'ORDER BY count(*) DESC;', (stats["max_activity_date"],)).fetchone()
        if max_activity_hour:
            stats["max_activity_hour"] = max_activity_hour[1]
        else:
            stats["max_activity_hour"] = None

    # Fleet size estimate: considering each line separately
    if max_activity_date and max_activity_hour:
        fleet_size_estimates = _fleet_size_estimate(gtfs, stats['max_activity_hour'], stats['max_activity_date'])
        stats.update(fleet_size_estimates)

    # Compute simple distributions of various columns that have a finite range of values.
    # Commented lines refer to values that are not imported yet, ?

    stats['routes__type__dist'] = _distribution(gtfs, 'routes', 'type')
    # stats['stop_times__pickup_type__dist'] = _distribution(gtfs, 'stop_times', 'pickup_type')
    # stats['stop_times__drop_off_type__dist'] = _distribution(gtfs, 'stop_times', 'drop_off_type')
    # stats['stop_times__timepoint__dist'] = _distribution(gtfs, 'stop_times', 'timepoint')
    stats['calendar_dates__exception_type__dist'] = _distribution(gtfs, 'calendar_dates', 'exception_type')
    stats['frequencies__exact_times__dist'] = _distribution(gtfs, 'frequencies', 'exact_times')
    stats['transfers__transfer_type__dist'] = _distribution(gtfs, 'transfers', 'transfer_type')
    stats['agencies__lang__dist'] = _distribution(gtfs, 'agencies', 'lang')
    stats['stops__location_type__dist'] = _distribution(gtfs, 'stops', 'location_type')
    # stats['stops__wheelchair_boarding__dist'] = _distribution(gtfs, 'stops', 'wheelchair_boarding')
    # stats['trips__wheelchair_accessible__dist'] = _distribution(gtfs, 'trips', 'wheelchair_accessible')
    # stats['trips__bikes_allowed__dist'] = _distribution(gtfs, 'trips', 'bikes_allowed')
    # stats[''] = _distribution(gtfs, '', '')
    stats = _feed_calendar_span(gtfs, stats)

    return stats


def _distribution(gtfs, table, column):
    """Count occurrences of values AND return it as a string.

    Example return value:   '1:5 2:15'"""
    cur = gtfs.conn.cursor()
    cur.execute('SELECT {column}, count(*) '
                'FROM {table} GROUP BY {column} '
                'ORDER BY {column}'.format(column=column, table=table))
    return ' '.join('%s:%s' % (t, c) for t, c in cur)


def _fleet_size_estimate(gtfs, hour, date):
    """
    Calculates fleet size estimates by two separate formula:
     1. Considering all routes separately with no interlining and doing a deficit calculation at every terminal
     2. By looking at the maximum number of vehicles in simultaneous movement

    Parameters
    ----------
    gtfs: GTFS
    hour: int
    date: ?

    Returns
    -------
    results: dict
        a dict with keys:
            fleet_size_route_based
            fleet_size_max_movement

    """
    results = {}

    fleet_size_list = []
    cur = gtfs.conn.cursor()
    rows = cur.execute(
        'SELECT type, max(vehicles) '
        'FROM ('
        'SELECT type, direction_id, sum(vehicles) as vehicles '
        'FROM '
        '('
        'SELECT trips.route_I, trips.direction_id, routes.route_id, name, type, count(*) as vehicles, cycle_time_min '
        'FROM trips, routes, days, '
        '('
        'SELECT first_trip.route_I, first_trip.direction_id, first_trip_start_time, first_trip_end_time, '
        'MIN(start_time_ds) as return_trip_start_time, end_time_ds as return_trip_end_time, '
        '(end_time_ds - first_trip_start_time)/60 as cycle_time_min '
        'FROM '
        'trips, '
        '(SELECT route_I, direction_id, MIN(start_time_ds) as first_trip_start_time, '
        'end_time_ds as first_trip_end_time '
        'FROM trips, days '
        'WHERE trips.trip_I=days.trip_I AND start_time_ds >= ? * 3600 '
        'AND start_time_ds <= (? + 1) * 3600 AND date = ? '
        'GROUP BY route_I, direction_id) first_trip '
        'WHERE first_trip.route_I = trips.route_I '
        'AND first_trip.direction_id != trips.direction_id '
        'AND start_time_ds >= first_trip_end_time '
        'GROUP BY trips.route_I, trips.direction_id'
        ') return_trip '
        'WHERE trips.trip_I=days.trip_I AND trips.route_I= routes.route_I '
        'AND date = ? AND trips.route_I = return_trip.route_I '
        'AND trips.direction_id = return_trip.direction_id '
        'AND start_time_ds >= first_trip_start_time '
        'AND start_time_ds < return_trip_end_time '
        'GROUP BY trips.route_I, trips.direction_id '
        'ORDER BY type, name, vehicles desc'
        ') cycle_times '
        'GROUP BY direction_id, type'
        ') vehicles_type '
        'GROUP BY type;', (hour, hour, date, date))
    for row in rows:
        fleet_size_list.append(str(row[0]) + ':' + str(row[1]))
    results['fleet_size_route_based'] = " ".join(fleet_size_list)

    # Fleet size estimate: maximum number of vehicles in movement
    fleet_size_list = []
    fleet_size_dict = {}
    if hour:
        for minute in range(hour * 3600, (hour + 1) * 3600, 60):
            rows = gtfs.conn.cursor().execute(
                'SELECT type, count(*) '
                'FROM trips, routes, days '
                'WHERE trips.route_I = routes.route_I '
                'AND trips.trip_I=days.trip_I '
                'AND start_time_ds <= ? '
                'AND end_time_ds > ? + 60 '
                'AND date = ? '
                'GROUP BY type;',
                (minute, minute, date))

            for row in rows:
                if fleet_size_dict.get(row[0], 0) < row[1]:
                    fleet_size_dict[row[0]] = row[1]

    for key in fleet_size_dict.keys():
        fleet_size_list.append(str(key) + ':' + str(fleet_size_dict[key]))
    results["fleet_size_max_movement"] = ' '.join(fleet_size_list)
    return results


def _n_gtfs_sources(gtfs):
    n_gtfs_sources = gtfs.execute_custom_query(
        "SELECT value FROM metadata WHERE key = 'n_gtfs_sources';").fetchone()
    if not n_gtfs_sources:
        n_gtfs_sources = [1]
    return n_gtfs_sources


def _feed_calendar_span(gtfs, stats):
    """
    Computes the temporal coverage of each source feed

    Parameters
    ----------
    gtfs: gtfspy.GTFS object
    stats: dict
        where to append the stats

    Returns
    -------
    stats: dict
    """
    n_feeds = _n_gtfs_sources(gtfs)[0]
    max_start = None
    min_end = None
    if n_feeds > 1:
        for i in range(n_feeds):
            feed_key = "feed_" + str(i) + "_"
            start_key = feed_key + "calendar_start"
            end_key = feed_key + "calendar_end"
            calendar_span = gtfs.conn.cursor().execute(
                'SELECT min(date), max(date) FROM trips, days '
                'WHERE trips.trip_I = days.trip_I AND trip_id LIKE ?;', (feed_key + '%',)).fetchone()

            stats[start_key] = calendar_span[0]
            stats[end_key] = calendar_span[1]
            if calendar_span[0] is not None and calendar_span[1] is not None:
                if not max_start and not min_end:
                    max_start = calendar_span[0]
                    min_end = calendar_span[1]
                else:
                    if gtfs.get_day_start_ut(calendar_span[0]) > gtfs.get_day_start_ut(max_start):
                        max_start = calendar_span[0]
                    if gtfs.get_day_start_ut(calendar_span[1]) < gtfs.get_day_start_ut(min_end):
                        min_end = calendar_span[1]
        stats["latest_feed_start_date"] = max_start
        stats["earliest_feed_end_date"] = min_end
    else:
        stats["latest_feed_start_date"] = stats["start_date"]
        stats["earliest_feed_end_date"] = stats["end_date"]
    return stats


def update_stats(gtfs):
    """
    Computes stats AND stores them into the underlying gtfs object (i.e. database).

    Parameters
    ----------
    gtfs: GTFS
    """
    stats = get_stats(gtfs)
    gtfs.update_stats(stats)


def trip_stats(gtfs, results_by_mode=False):
    """

    Parameters
    ----------
    gtfs: GTFS
    results_by_mode: bool

    Returns
    -------
    if results_by_mode is False:
        q_result: pandas.DataFrame
    if results_by_mode is True:
        q_results: dict
            a dict with the following keys:
                [ADD HERE]
    """
    conn = gtfs.conn

    conn.create_function("find_distance", 4, wgs84_distance)
    cur = conn.cursor()
    # this query calculates the distance and travel time for each complete trip
    # stop_data_df = pd.read_sql_query(query, self.conn, params=params)

    query = 'SELECT ' \
            'startstop.trip_I AS trip_I, ' \
            'type, ' \
            'sum(CAST(find_distance(startstop.lat, startstop.lon, endstop.lat, endstop.lon) AS INT)) as total_distance, ' \
            'sum(endstop.arr_time_ds - startstop.arr_time_ds) as total_traveltime ' \
            'FROM ' \
            '(SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) startstop, ' \
            '(SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) endstop, ' \
            'trips, ' \
            'routes ' \
            'WHERE ' \
            'startstop.trip_I = endstop.trip_I ' \
            'AND startstop.seq + 1 = endstop.seq ' \
            'AND startstop.trip_I = trips.trip_I ' \
            'AND trips.route_I = routes.route_I ' \
            'GROUP BY startstop.trip_I'

    q_result = pd.read_sql_query(query, conn)
    q_result['avg_speed_kmh'] = 3.6 * q_result['total_distance'] / q_result['total_traveltime']
    q_result['total_distance'] = q_result['total_distance'] / 1000
    q_result['total_traveltime'] = q_result['total_traveltime'] / 60
    q_result = q_result.loc[q_result['avg_speed_kmh'] != float("inf")]

    if results_by_mode:
        q_results = {}
        for type in q_result['type'].unique().tolist():
            q_results[type] = q_result.loc[q_result['type'] == type]
        return q_results
    else:
        return q_result


def get_section_stats(gtfs, results_by_mode=False):
    conn = gtfs.conn

    conn.create_function("find_distance", 4, wgs84_distance)
    cur = conn.cursor()
    # this query calculates the distance and travel time for each stop to stop section for each trip
    # stop_data_df = pd.read_sql_query(query, self.conn, params=params)

    query = 'SELECT type, from_stop_I, to_stop_I, distance, min(travel_time) AS min_time, max(travel_time) AS max_time, avg(travel_time) AS mean_time ' \
            'FROM ' \
            '(SELECT q1.trip_I, type, q1.stop_I as from_stop_I, q2.stop_I as to_stop_I,  ' \
            'CAST(find_distance(q1.lat, q1.lon, q2.lat, q2.lon) AS INT) as distance, ' \
            'q2.arr_time_ds - q1.arr_time_ds as travel_time, ' \
            'q1.lat AS from_lat, q1.lon AS from_lon, q2.lat AS to_lat, q2.lon AS to_lon ' \
            'FROM ' \
            '(SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) q1, ' \
            '(SELECT * FROM stop_times, stops WHERE stop_times.stop_I = stops.stop_I) q2, ' \
            'trips, ' \
            'routes ' \
            'WHERE q1.trip_I = q2.trip_I ' \
            'AND q1.seq + 1 = q2.seq ' \
            'AND q1.trip_I = trips.trip_I ' \
            'AND trips.route_I = routes.route_I) sq1 ' \
            'GROUP BY to_stop_I, from_stop_I, type '

    q_result = pd.read_sql_query(query, conn)

    if results_by_mode:
        q_results = {}
        for type in q_result['type'].unique().tolist():
            q_results[type] = q_result.loc[q_result['type'] == type]
        return q_results
    else:
        return q_result


def route_frequencies(gtfs, results_by_mode=False):
    """
    Return the frequency of all types of routes per day.

    Parameters
    -----------
    gtfs: GTFS

    Returns
    -------
    pandas.DataFrame with columns
        route_I, type, frequency
    """
    day = gtfs.get_suitable_date_for_daily_extract()
    query = (
        " SELECT f.route_I, type, frequency FROM routes as r"
        " JOIN"
        " (SELECT route_I, COUNT(route_I) as frequency"
        " FROM"
        " (SELECT date, route_I, trip_I"
        " FROM day_stop_times"
        " WHERE date = '{day}'"
        " GROUP by route_I, trip_I)"
        " GROUP BY route_I) as f"
        " ON f.route_I = r.route_I"
        " ORDER BY frequency DESC".format(day=day))
    
    return pd.DataFrame(gtfs.execute_custom_query_pandas(query))


def hourly_frequencies(gtfs, st, et, route_type):
    """
    Return all the number of vehicles (i.e. busses,trams,etc) that pass hourly through a stop in a time frame.

    Parameters
    ----------
    gtfs: GTFS
    st : int
        start time of the time framein unix time
    et : int
        end time of the time frame in unix time
    route_type: int

    Returns
    -------
    numeric pandas.DataFrame with columns
        stop_I, lat, lon, frequency
    """ 
    timeframe = et-st
    hours = timeframe/ 3600
    day = gtfs.get_suitable_date_for_daily_extract()
    stops = gtfs.get_stops_for_route_type(route_type).T.drop_duplicates().T
    query = ("SELECT * FROM stops as x"
             " JOIN"
             " (SELECT * , COUNT(*)/{h} as frequency"
             " FROM stop_times, days"
             " WHERE stop_times.trip_I = days.trip_I"
             " AND dep_time_ds > {st}"
             " AND dep_time_ds < {et}"
             " AND date = '{day}'"
             " GROUP BY stop_I) as y"
             " ON y.stop_I = x.stop_I".format(h=hours, st=st, et=et, day=day))
    try:
        trips_frequency = gtfs.execute_custom_query_pandas(query).T.drop_duplicates().T
        df = pd.merge(stops[['stop_I', 'lat', 'lon']], trips_frequency[['stop_I', 'frequency']],
                      on='stop_I', how='inner')
        return df.apply(pd.to_numeric)
    except:
        raise ValueError("Maybe too short time frame!")


def frequencies_by_generated_route(gtfs, st, et, day=None):
    timeframe = et-st
    hours = timeframe/3600
    if not day:
        day = gtfs.get_suitable_date_for_daily_extract()
    query = """SELECT count(*)/{h} AS frequency, count(*) AS n_trips, route, type FROM 
    (SELECT trip_I, group_concat(stop_I) AS route, name, type FROM
    (SELECT * FROM stop_times, days, trips, routes
    WHERE stop_times.trip_I = days.trip_I AND stop_times.trip_I = trips.trip_I AND  routes.route_I = trips.route_I AND 
    days.date = '{day}' AND start_time_ds >= {st} AND start_time_ds < {et}
    ORDER BY trip_I, seq) q1
    GROUP BY trip_I) q2
    GROUP BY route""".format(h=hours, st=st, et=et, day=day)
    df = gtfs.execute_custom_query_pandas(query)
    return df


def departure_stops(gtfs, st, et):
    day = gtfs.get_suitable_date_for_daily_extract()
    query = """select stop_I, count(*) as n_departures from 
                (select min(seq), * from stop_times, days, trips
                where stop_times.trip_I = days.trip_I and stop_times.trip_I = trips.trip_I and days.date = '{day}'
                 and start_time_ds >= {st} and start_time_ds < {et}
                group by stop_times.trip_I) q1
                group by stop_I""".format(st=st, et=et, day=day)
    df = gtfs.execute_custom_query_pandas(query)
    df = gtfs.add_coordinates_to_df(df)
    return df


def get_vehicle_hours_by_type(gtfs, route_type):
    """
    Return the sum of vehicle hours in a particular day by route type.
    """

    day = gtfs.get_suitable_date_for_daily_extract()
    query = (" SELECT * , SUM(end_time_ds - start_time_ds)/3600 as vehicle_hours_type"
             " FROM"
             " (SELECT * FROM day_trips as q1"
             " INNER JOIN"
             " (SELECT route_I, type FROM routes) as q2"
             " ON q1.route_I = q2.route_I"
             " WHERE type = {route_type}"
             " AND date = '{day}')".format(day=day, route_type=route_type))
    df = gtfs.execute_custom_query_pandas(query)
    return df['vehicle_hours_type'].item()


def trips_frequencies(gtfs):
    """
        Get the frequency of trip_I in a particular day
    """
    query = (
        " SELECT q1.stop_I as from_stop_I, q2.stop_I as to_stop_I, q1.trip_I as trip_I, COUNT(*) as freq FROM"
        " (SELECT * FROM stop_times) q1,"
        " (SELECT * FROM stop_times) q2"
        " WHERE q1.seq+1=q2.seq AND q1.trip_I=q2.trip_I"
        " GROUP BY from_stop_I, to_stop_I")
    return(gtfs.execute_custom_query_pandas(query))

# def route_circuity():
#    pass



import copy

from gtfspy.routing.label import compute_pareto_front
from gtfspy.routing.node_profile_analyzer_time import NodeProfileAnalyzerTime
from gtfspy.routing.profile_block_analyzer import ProfileBlockAnalyzer
from gtfspy.routing.profile_block import ProfileBlock


class FastestPathAnalyzer:

    def __init__(self, labels, start_time_dep, end_time_dep, walk_duration=float('inf'), label_props_to_consider=None, **kwargs):
        """
        Parameters
        ----------
        labels: list
            List of labels (each label should at least have attributes "departure_time" and "arrival_time")
        walk_duration: float
            What is the maximum duration for a journey to be considered.
        label_props_to_consider: list
        """
        for label in labels:
            assert (hasattr(label, "departure_time"))
            assert (hasattr(label, "arrival_time_target"))
        self.start_time_dep = start_time_dep
        self.end_time_dep = end_time_dep
        self.walk_duration = walk_duration
        if label_props_to_consider is None:
            self.label_props = []
        else:
            self.label_props = label_props_to_consider
        self._fastest_path_labels = self._compute_fastest_path_labels(labels)
        # assert each label has the required properties
        for label in self._fastest_path_labels:
            for prop in self.label_props:
                assert (hasattr(label, prop))
        self.kwargs = kwargs

    def _compute_fastest_path_labels(self, labels):
        relevant_labels = [label.get_copy() for label in labels if (self.start_time_dep < label.departure_time <= self.end_time_dep)]
        if len(relevant_labels) is 0 or relevant_labels[-1].departure_time < self.end_time_dep:
            # add an after label
            smallest_arr_time_after_end_time = float('inf')
            smallest_arr_time_label = None
            for label in labels:
                if self.end_time_dep < label.departure_time and (label.arrival_time_target < smallest_arr_time_after_end_time):
                    smallest_arr_time_after_end_time = label.arrival_time_target
                    smallest_arr_time_label = label
            if smallest_arr_time_label is not None:
                relevant_labels.append(smallest_arr_time_label.get_copy())

        for label in relevant_labels:
            if hasattr(label, "first_leg_is_walk"):
                label.first_leg_is_walk = False
        fp_labels = list(reversed(compute_pareto_front(relevant_labels, ignore_n_boardings=True)))

        # assert ordered:
        for i in range(len(fp_labels) - 1):
            try:
                assert (fp_labels[i].arrival_time_target <= fp_labels[i + 1].arrival_time_target)
                assert (fp_labels[i].departure_time < fp_labels[i + 1].departure_time)
            except AssertionError as e:
                for fp_label in fp_labels:
                    print(fp_label)
                print(fp_labels[i].arrival_time_target, fp_labels[i + 1].arrival_time_target)
                print(fp_labels[i].departure_time, fp_labels[i + 1].departure_time)
                raise e
        return fp_labels

    def get_fastest_path_labels(self, include_next_label_outside_interval=False):
        if include_next_label_outside_interval or not self._fastest_path_labels:
            return self._fastest_path_labels
        else:
            if self._fastest_path_labels[-1].departure_time == self.end_time_dep:
                return self._fastest_path_labels
            else:
                return self._fastest_path_labels[:-1]

    def calculate_pre_journey_waiting_times_ignoring_direct_walk(self):
        previous_label = None
        for label in self._fastest_path_labels:
            if previous_label:
                label.pre_journey_wait_fp = label.departure_time - previous_label.departure_time
            else:
                label.pre_journey_wait_fp = label.departure_time - self.start_time_dep
            previous_label = label

    def get_fastest_path_temporal_distance_blocks(self):
        """
        Returns
        -------
        blocks: list[ProfileBlock]
        """
        def _label_to_prop_dict(label):
            return {prop: getattr(label, prop) for prop in self.label_props}

        labels = self._fastest_path_labels
        for i in range(len(labels) - 1):
            assert (labels[i].departure_time < labels[i + 1].departure_time)

        previous_dep_time = self.start_time_dep
        blocks = []
        for label in labels:
            if previous_dep_time >= self.end_time_dep:
                break
            end_time = min(label.departure_time, self.end_time_dep)
            assert (end_time >= previous_dep_time)

            temporal_distance_start = label.duration() + (label.departure_time - previous_dep_time)

            if temporal_distance_start > self.walk_duration:
                split_point_x_computed = label.departure_time - (self.walk_duration - label.duration())
                split_point_x = min(split_point_x_computed, end_time)
                if previous_dep_time < split_point_x:
                    # add walk block, only if it is required
                    walk_block = ProfileBlock(previous_dep_time,
                                              split_point_x,
                                              self.walk_duration,
                                              self.walk_duration,
                                              **_label_to_prop_dict(label))
                    blocks.append(walk_block)
                if split_point_x < end_time:
                    trip_block = ProfileBlock(split_point_x, end_time,
                                              label.duration() + (end_time - split_point_x),
                                              label.duration(),
                                              **_label_to_prop_dict(label))
                    blocks.append(trip_block)
            else:
                journey_block = ProfileBlock(
                    previous_dep_time,
                    end_time,
                    temporal_distance_start,
                    temporal_distance_start - (end_time - previous_dep_time),
                    **_label_to_prop_dict(label))
                blocks.append(journey_block)
            previous_dep_time = blocks[-1].end_time

        if previous_dep_time < self.end_time_dep:
            last_block = ProfileBlock(previous_dep_time,
                                      self.end_time_dep,
                                      self.walk_duration,
                                      self.walk_duration)
            blocks.append(last_block)
        return blocks

    def get_time_analyzer(self):
        """
        Returns
        -------
        NodeProfileAnalyzerTime
        """
        return NodeProfileAnalyzerTime(self._fastest_path_labels,
                                       self.walk_duration,
                                       self.start_time_dep,
                                       self.end_time_dep)

    def get_props(self):
        return list(self.label_props)

    def get_temporal_distance_analyzer(self):
        kwargs = self.kwargs
        return ProfileBlockAnalyzer(self.get_fastest_path_temporal_distance_blocks(), **kwargs)

    def get_prop_analyzer_for_pre_journey_wait(self):
        kwargs = self.kwargs
        prop_blocks = []
        fp_blocks = self.get_fastest_path_temporal_distance_blocks()
        for b in fp_blocks:
            if b.distance_end == float("inf"):
                prop_block = b
            elif b.is_flat():
                prop_block = ProfileBlock(b.start_time, b.end_time, 0, 0)
            else:
                prop_block = ProfileBlock(b.start_time, b.end_time, b.width(), 0)
            prop_blocks.append(prop_block)
        return ProfileBlockAnalyzer(prop_blocks, **kwargs)

    def get_prop_analyzer_flat(self, property, value_no_next_journey, value_cutoff):
        """
        Get a journey property analyzer, where each journey is weighted by the number of.

        Parameters
        ----------
        property: string
            Name of the property, needs to be one of label_props given on initialization.
        value_no_next_journey:
            Value of the profile, when there is no next journey available.
        value_cutoff: number
            default value of the property when cutoff is applied

        Returns
        -------
        ProfileBlockAnalyzer
        """
        kwargs = self.kwargs
        fp_blocks = self.get_fastest_path_temporal_distance_blocks()
        prop_blocks = []
        for b in fp_blocks:
            if b.is_flat():
                if b.distance_end == self.walk_duration and b.distance_end != float('inf'):
                    prop_value = value_cutoff
                else:
                    prop_value = value_no_next_journey
            else:
                prop_value = b[property]
            prop_block = ProfileBlock(b.start_time, b.end_time, prop_value, prop_value)
            prop_blocks.append(prop_block)
        return ProfileBlockAnalyzer(prop_blocks, **kwargs)






import io
from unittest import TestCase

from gtfspy.warnings_container import WarningsContainer


class TestWarningsContainer(TestCase):

    def test_summary_print(self):
        wc = WarningsContainer()
        wc.add_warning("DUMMY_WARNING", ["dummy1", "dummy2"], 2)

        f = io.StringIO("")
        wc.write_summary(output_stream=f)
        f.seek(0)
        assert len(f.readlines()) == len(wc.get_warning_counter().keys()) + 1

    def test_details_print(self):
        wc = WarningsContainer()
        wc.add_warning("DUMMY_WARNING", ["dummy1", "dummy2"], 2)

        f = io.StringIO("")
        wc.write_details(output_stream=f)
        f.seek(0)
        assert len(f.readlines()) > len(wc.get_warning_counter().keys()) + 1


from unittest import TestCase

from gtfspy.routing.label import LabelTime
from gtfspy.routing.node_profile_c import NodeProfileC


class TestNodeProfileC(TestCase):

    def test_earliest_arrival_time(self):
        node_profile = NodeProfileC()
        self.assertEquals(float("inf"), node_profile.evaluate_earliest_arrival_time_at_target(0, 0))

        node_profile.update_pareto_optimal_tuples(LabelTime(departure_time=3, arrival_time_target=4))
        self.assertEquals(4, node_profile.evaluate_earliest_arrival_time_at_target(2, 0))

        node_profile.update_pareto_optimal_tuples(LabelTime(departure_time=1, arrival_time_target=1))
        self.assertEquals(1, node_profile.evaluate_earliest_arrival_time_at_target(0, 0))

    def test_pareto_optimality(self):
        node_profile = NodeProfileC()

        pair3 = LabelTime(departure_time=2, arrival_time_target=3)
        node_profile.update_pareto_optimal_tuples(pair3)
        pair2 = LabelTime(departure_time=1, arrival_time_target=2)
        node_profile.update_pareto_optimal_tuples(pair2)
        pair1 = LabelTime(departure_time=1, arrival_time_target=1)
        node_profile.update_pareto_optimal_tuples(pair1)
        self.assertEqual(2, len(node_profile.get_final_optimal_labels()))

    def test_pareto_optimality2(self):
        node_profile = NodeProfileC()
        pt2 = LabelTime(departure_time=10, arrival_time_target=35)
        node_profile.update_pareto_optimal_tuples(pt2)
        pt1 = LabelTime(departure_time=5, arrival_time_target=35)
        node_profile.update_pareto_optimal_tuples(pt1)
        print(node_profile._labels)
        self.assertEquals(len(node_profile.get_final_optimal_labels()), 1)

    def test_identity_profile(self):
        identity_profile = NodeProfileC(0)
        self.assertFalse(identity_profile.update_pareto_optimal_tuples(LabelTime(10, 10)))
        self.assertEqual(10, identity_profile.evaluate_earliest_arrival_time_at_target(10, 0))

    def test_walk_duration(self):
        node_profile = NodeProfileC(walk_to_target_duration=27)
        self.assertEqual(27, node_profile.get_walk_to_target_duration())
        pt2 = LabelTime(departure_time=10, arrival_time_target=35)
        pt1 = LabelTime(departure_time=5, arrival_time_target=35)
        node_profile.update_pareto_optimal_tuples(pt2)
        node_profile.update_pareto_optimal_tuples(pt1)
        self.assertEqual(1, len(node_profile.get_final_optimal_labels()))

    def test_assert_raises_wrong_order(self):
        node_profile = NodeProfileC()
        pt1 = LabelTime(departure_time=5, arrival_time_target=35)
        pt2 = LabelTime(departure_time=10, arrival_time_target=35)
        self.assertTrue(node_profile.update_pareto_optimal_tuples(pt1))
        with self.assertRaises(AssertionError):
            node_profile.update_pareto_optimal_tuples(pt2)



import os
import unittest

from gtfspy.gtfs import GTFS
from gtfspy.timetable_validator import TimetableValidator

class TestGTFSValidator(unittest.TestCase):

    def setUp(self):
        self.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        self.G = GTFS.from_directory_as_inmemory_db(self.gtfs_source_dir)

    def test_compiles(self):
        validator = TimetableValidator(self.G)
        warnings = validator.validate_and_get_warnings()
        warning_counts = warnings.get_warning_counter()
        assert len(warning_counts) > 0





import datetime
import warnings

import numpy
import pytz

from matplotlib import dates as md, rcParams
import matplotlib.pyplot as plt

from gtfspy.routing.profile_block_analyzer import ProfileBlockAnalyzer
from gtfspy.routing.profile_block import ProfileBlock
from gtfspy.routing.node_profile_simple import NodeProfileSimple


def _if_no_trips_return_inf(func):
    def wrapper(self):
        if self.trip_durations:
            return func(self)
        else:
            return float('inf')

    return wrapper


class NodeProfileAnalyzerTime:

    @classmethod
    def from_profile(cls, node_profile, start_time_dep, end_time_dep):
        assert isinstance(node_profile, NodeProfileSimple), type(node_profile)
        return NodeProfileAnalyzerTime(node_profile.get_final_optimal_labels(),
                                       node_profile.get_walk_to_target_duration(),
                                       start_time_dep,
                                       end_time_dep)

    def __init__(self, labels, walk_time_to_target, start_time_dep, end_time_dep):
        """
        Initialize the data structures required by

        Parameters
        ----------
        node_profile: NodeProfileSimple

        """
        self.start_time_dep = start_time_dep
        self.end_time_dep = end_time_dep
        # used for computing temporal distances:
        all_pareto_optimal_tuples = [pt for pt in labels if
                                      (start_time_dep < pt.departure_time < end_time_dep)]

        labels_after_dep_time = [label for label in labels if label.departure_time >= self.end_time_dep]
        if labels_after_dep_time:
            next_label_after_end_time = min(labels_after_dep_time, key=lambda el: el.arrival_time_target)
            all_pareto_optimal_tuples.append(next_label_after_end_time)

        all_pareto_optimal_tuples = sorted(all_pareto_optimal_tuples, key=lambda ptuple: ptuple.departure_time)

        arrival_time_target_at_end_time = end_time_dep + walk_time_to_target
        previous_trip = None
        for trip_tuple in all_pareto_optimal_tuples:
            if previous_trip:
                assert(trip_tuple.arrival_time_target > previous_trip.arrival_time_target)
            if trip_tuple.departure_time > self.end_time_dep \
                    and trip_tuple.arrival_time_target < arrival_time_target_at_end_time:
                arrival_time_target_at_end_time = trip_tuple.arrival_time_target
            previous_trip = trip_tuple

        self._walk_time_to_target = walk_time_to_target
        self._profile_blocks = []
        previous_departure_time = start_time_dep
        self.trip_durations = []
        self.trip_departure_times = []
        for trip_pareto_tuple in all_pareto_optimal_tuples:
            if trip_pareto_tuple.departure_time > self.end_time_dep:
                continue
            if self._walk_time_to_target <= trip_pareto_tuple.duration():
                print(self._walk_time_to_target, trip_pareto_tuple.duration())
                assert(self._walk_time_to_target > trip_pareto_tuple.duration())
            effective_trip_previous_departure_time = max(
                previous_departure_time,
                trip_pareto_tuple.departure_time - (self._walk_time_to_target - trip_pareto_tuple.duration())
            )
            if effective_trip_previous_departure_time > previous_departure_time:
                walk_block = ProfileBlock(start_time=previous_departure_time,
                                          end_time=effective_trip_previous_departure_time,
                                          distance_start=self._walk_time_to_target,
                                          distance_end=self._walk_time_to_target
                                          )
                self._profile_blocks.append(walk_block)
            trip_waiting_time = trip_pareto_tuple.departure_time - effective_trip_previous_departure_time
            trip_block = ProfileBlock(end_time=trip_pareto_tuple.departure_time,
                                      start_time=effective_trip_previous_departure_time,
                                      distance_start=trip_pareto_tuple.duration() + trip_waiting_time,
                                      distance_end=trip_pareto_tuple.duration())
            self.trip_durations.append(trip_pareto_tuple.duration())
            self.trip_departure_times.append(trip_pareto_tuple.departure_time)
            self._profile_blocks.append(trip_block)
            previous_departure_time = trip_pareto_tuple.departure_time

        # deal with last (or add walking block like above)
        if not self._profile_blocks or self._profile_blocks[-1].end_time < end_time_dep:
            if len(self._profile_blocks) > 0:
                dep_previous = self._profile_blocks[-1].end_time
            else:
                dep_previous = start_time_dep
            waiting_time = end_time_dep - dep_previous

            distance_end_trip = arrival_time_target_at_end_time - end_time_dep
            walking_wait_time = min(end_time_dep - dep_previous,
                                    waiting_time - (self._walk_time_to_target - distance_end_trip))
            walking_wait_time = max(0, walking_wait_time)
            if walking_wait_time > 0:
                walk_block = ProfileBlock(start_time=dep_previous,
                                          end_time=dep_previous + walking_wait_time,
                                          distance_start=self._walk_time_to_target,
                                          distance_end=self._walk_time_to_target
                                          )
                assert (walk_block.start_time <= walk_block.end_time)
                assert (walk_block.distance_end <= walk_block.distance_start)
                self._profile_blocks.append(walk_block)
            trip_waiting_time = waiting_time - walking_wait_time

            if trip_waiting_time > 0:
                try:
                    trip_block = ProfileBlock(start_time=dep_previous + walking_wait_time,
                                          end_time=dep_previous + walking_wait_time + trip_waiting_time,
                                          distance_start=distance_end_trip + trip_waiting_time,
                                          distance_end=distance_end_trip)
                    assert (trip_block.start_time <= trip_block.end_time)
                    assert (trip_block.distance_end <= trip_block.distance_start)
                    self._profile_blocks.append(trip_block)
                except AssertionError as e:
                    # the error was due to a very small waiting timesmall numbers
                    assert(trip_waiting_time < 10**-5)

        # TODO? Refactor to use the cutoff_distance feature in ProfileBlockAnalyzer?
        self.profile_block_analyzer = ProfileBlockAnalyzer(profile_blocks=self._profile_blocks)

    def n_pareto_optimal_trips(self):
        """
        Get number of pareto-optimal trips

        Returns
        -------
        n_trips: float
        """
        return float(len(self.trip_durations))

    @_if_no_trips_return_inf
    def min_trip_duration(self):
        """
        Get minimum travel time to destination.

        Returns
        -------
        float: min_trip_duration
            float('nan') if no trips take place
        """
        return numpy.min(self.trip_durations)

    @_if_no_trips_return_inf
    def max_trip_duration(self):
        """
        Get minimum travel time to destination.

        Returns
        -------
        float: max_trip_duration
            float('inf') if no trips take place
        """
        return numpy.max(self.trip_durations)

    @_if_no_trips_return_inf
    def mean_trip_duration(self):
        """
        Get average travel time to destination.

        Returns
        -------
        float: max_trip_duration
            float('inf') if no trips take place
        """
        return numpy.mean(self.trip_durations)

    @_if_no_trips_return_inf
    def median_trip_duration(self):
        """
        Get average travel time to destination.

        Returns
        -------
        float: max_trip_duration
            float('inf') if no trips take place
        """
        return numpy.median(self.trip_durations)

    def mean_temporal_distance(self):
        """
        Get mean temporal distance (in seconds) to the target.

        Returns
        -------
        mean_temporal_distance : float
        """
        total_width = self.end_time_dep - self.start_time_dep
        total_area = sum([block.area() for block in self._profile_blocks])
        return total_area / total_width

    def median_temporal_distance(self):
        """
        Returns
        -------
        median_temporal_distance : float
        """
        return self.profile_block_analyzer.median()

    def min_temporal_distance(self):
        """
        Compute the minimum temporal distance to target.

        Returns
        -------
        min_temporal_distance: float
        """
        return self.profile_block_analyzer.min()

    def max_temporal_distance(self):
        """
        Compute the maximum temporal distance.

        Returns
        -------
        max_temporal_distance : float
        """
        return self.profile_block_analyzer.max()

    def largest_finite_temporal_distance(self):
        """
        Compute the maximum temporal distance.

        Returns
        -------
        max_temporal_distance : float
        """
        return self.profile_block_analyzer.largest_finite_distance()

    def plot_temporal_distance_cdf(self):
        """
        Plot the temporal distance cumulative density function.

        Returns
        -------
        fig: matplotlib.Figure
        """
        xvalues, cdf = self.profile_block_analyzer._temporal_distance_cdf()
        fig = plt.figure()
        ax = fig.add_subplot(111)
        xvalues = numpy.array(xvalues) / 60.0
        ax.plot(xvalues, cdf, "-k")
        ax.fill_between(xvalues, cdf, color="red", alpha=0.2)
        ax.set_ylabel("CDF(t)")
        ax.set_xlabel("Temporal distance t (min)")
        return fig

    def plot_temporal_distance_pdf(self, use_minutes=True, color="green", ax=None):
        """
        Plot the temporal distance probability density function.

        Returns
        -------
        fig: matplotlib.Figure
        """
        from matplotlib import pyplot as plt
        plt.rc('text', usetex=True)
        temporal_distance_split_points_ordered, densities, delta_peaks = self._temporal_distance_pdf()
        xs = []
        for i, x in enumerate(temporal_distance_split_points_ordered):
            xs.append(x)
            xs.append(x)
        xs = numpy.array(xs)
        ys = [0]
        for y in densities:
            ys.append(y)
            ys.append(y)
        ys.append(0)
        ys = numpy.array(ys)
        # convert data to minutes:
        xlabel = "Temporal distance (s)"
        ylabel = "Probability density (t)"
        if use_minutes:
            xs /= 60.0
            ys *= 60.0
            xlabel = "Temporal distance (min)"
            delta_peaks = {peak / 60.0: mass for peak, mass in delta_peaks.items()}

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)
        ax.plot(xs, ys, "k-")
        ax.fill_between(xs, ys, color="green", alpha=0.2)

        if delta_peaks:
            peak_height = max(ys) * 1.4
            max_x = max(xs)
            min_x = min(xs)
            now_max_x = max(xs) + 0.3 * (max_x - min_x)
            now_min_x = min_x - 0.1 * (max_x - min_x)

            text_x_offset = 0.1 * (now_max_x - max_x)

            for loc, mass in delta_peaks.items():
                ax.plot([loc, loc], [0, peak_height], color="green", lw=5)
                ax.text(loc + text_x_offset, peak_height * 0.99, "$P(\\mathrm{walk}) = %.2f$" % (mass), color="green")
            ax.set_xlim(now_min_x, now_max_x)

            tot_delta_peak_mass = sum(delta_peaks.values())
            transit_text_x = (min_x + max_x) / 2
            transit_text_y = min(ys[ys > 0]) / 2.
            ax.text(transit_text_x,
                    transit_text_y,
                    "$P(mathrm{PT}) = %.2f$" % (1 - tot_delta_peak_mass),
                    color="green",
                    va="center",
                    ha="center")

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.set_ylim(bottom=0)
        return ax.figure

    def plot_temporal_distance_pdf_horizontal(self, use_minutes=True,
                                              color="green",
                                              ax=None,
                                              duration_divider=60.0,
                                              legend_font_size=None,
                                              legend_loc=None):
        """
        Plot the temporal distance probability density function.

        Returns
        -------
        fig: matplotlib.Figure
        """
        from matplotlib import pyplot as plt
        plt.rc('text', usetex=True)

        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        temporal_distance_split_points_ordered, densities, delta_peaks = self._temporal_distance_pdf()
        xs = []
        for i, x in enumerate(temporal_distance_split_points_ordered):
            xs.append(x)
            xs.append(x)
        xs = numpy.array(xs)
        ys = [0]
        for y in densities:
            ys.append(y)
            ys.append(y)
        ys.append(0)
        ys = numpy.array(ys)
        # convert data to minutes:
        xlabel = "Temporal distance (s)"
        ylabel = "Probability density $P(\\tau)$"
        if use_minutes:
            xs /= duration_divider
            ys *= duration_divider
            xlabel = "Temporal distance (min)"
            delta_peaks = {peak / 60.0: mass for peak, mass in delta_peaks.items()}

        if delta_peaks:
            peak_height = max(ys) * 1.4
            max_x = max(xs)
            min_x = min(xs)
            now_max_x = max(xs) + 0.3 * (max_x - min_x)
            now_min_x = min_x - 0.1 * (max_x - min_x)

            text_x_offset = 0.1 * (now_max_x - max_x)

            for loc, mass in delta_peaks.items():
                text = "$P(\\mathrm{walk}) = " + ("%.2f$" % (mass))
                ax.plot([0, peak_height], [loc, loc], color=color, lw=5, label=text)

        ax.plot(ys, xs, "k-")
        if delta_peaks:
            tot_delta_peak_mass = sum(delta_peaks.values())
            fill_label = "$P(\\mathrm{PT}) = %.2f$" % (1-tot_delta_peak_mass)
        else:
            fill_label = None
        ax.fill_betweenx(xs, ys, color=color, alpha=0.2, label=fill_label)

        ax.set_ylabel(xlabel)
        ax.set_xlabel(ylabel)
        ax.set_xlim(left=0, right=max(ys) * 1.2)
        if delta_peaks:
            if legend_font_size is None:
                legend_font_size = 12
            if legend_loc is None:
                legend_loc = "best"
            ax.legend(loc=legend_loc, prop={'size': legend_font_size})


        if True:
            line_tyles = ["-.", "--", "-"][::-1]
            to_plot_funcs = [self.max_temporal_distance, self.mean_temporal_distance, self.min_temporal_distance]

            xmin, xmax = ax.get_xlim()
            for to_plot_func, ls in zip(to_plot_funcs, line_tyles):
                y = to_plot_func() / duration_divider
                assert y < float('inf')
                # factor of 10 just to be safe that the lines cover the whole region.
                ax.plot([xmin, xmax*10], [y, y], color="black", ls=ls, lw=1)

        return ax.figure

    def plot_temporal_distance_profile(self,
                                       timezone=None,
                                       color="black",
                                       alpha=0.15,
                                       ax=None,
                                       lw=2,
                                       label="",
                                       plot_tdist_stats=False,
                                       plot_trip_stats=False,
                                       format_string="%Y-%m-%d %H:%M:%S",
                                       plot_journeys=False,
                                       duration_divider=60.0,
                                       fill_color="green",
                                       journey_letters=None,
                                       return_letters=False):
        """
        Parameters
        ----------
        timezone: str
        color: color
        format_string: str, None
            if None, the original values are used
        plot_journeys: bool, optional
            if True, small dots are plotted at the departure times
        """
        if ax is None:
            fig = plt.figure()
            ax = fig.add_subplot(111)

        if timezone is None:
            warnings.warn("Warning: No timezone specified, defaulting to UTC")
            timezone = pytz.timezone("Etc/UTC")

        def _ut_to_unloc_datetime(ut):
            dt = datetime.datetime.fromtimestamp(ut, timezone)
            return dt.replace(tzinfo=None)

        if format_string:
            x_axis_formatter = md.DateFormatter(format_string)
            ax.xaxis.set_major_formatter(x_axis_formatter)
        else:
            _ut_to_unloc_datetime = lambda x: x

        ax.set_xlim(
            _ut_to_unloc_datetime(self.start_time_dep),
            _ut_to_unloc_datetime(self.end_time_dep)
        )

        if plot_tdist_stats:
            line_tyles = ["-.", "--", "-"][::-1]
            # to_plot_labels = ["maximum temporal distance", "mean temporal distance", "minimum temporal distance"]
            to_plot_labels  = ["$\\tau_\\mathrm{max} \\;$ = ", "$\\tau_\\mathrm{mean}$ = ", "$\\tau_\\mathrm{min} \\:\\:$ = "]
            to_plot_funcs = [self.max_temporal_distance, self.mean_temporal_distance, self.min_temporal_distance]

            xmin, xmax = ax.get_xlim()
            for to_plot_label, to_plot_func, ls in zip(to_plot_labels, to_plot_funcs, line_tyles):
                y = to_plot_func() / duration_divider
                assert y < float('inf'), to_plot_label
                to_plot_label = to_plot_label + "%.1f min" % (y)
                ax.plot([xmin, xmax], [y, y], color="black", ls=ls, lw=1, label=to_plot_label)

        if plot_trip_stats:
            assert (not plot_tdist_stats)
            line_tyles = ["-", "-.", "--"]
            to_plot_labels = ["min journey duration", "max journey duration", "mean journey duration"]
            to_plot_funcs = [self.min_trip_duration, self.max_trip_duration, self.mean_trip_duration]

            xmin, xmax = ax.get_xlim()
            for to_plot_label, to_plot_func, ls in zip(to_plot_labels, to_plot_funcs, line_tyles):
                y = to_plot_func() / duration_divider
                if not numpy.math.isnan(y):
                    ax.plot([xmin, xmax], [y, y], color="red", ls=ls, lw=2)
                    txt = to_plot_label + "\n = %.1f min" % y
                    ax.text(xmax + 0.01 * (xmax - xmin), y, txt, color="red", va="center", ha="left")

            old_xmax = xmax
            xmax += (xmax - xmin) * 0.3
            ymin, ymax = ax.get_ylim()
            ax.fill_between([old_xmax, xmax], ymin, ymax, color="gray", alpha=0.1)
            ax.set_xlim(xmin, xmax)

        # plot the actual profile
        vertical_lines, slopes = self.profile_block_analyzer.get_vlines_and_slopes_for_plotting()
        for i, line in enumerate(slopes):
            xs = [_ut_to_unloc_datetime(x) for x in line['x']]
            if i is 0:
                label = u"profile"
            else:
                label = None
            ax.plot(xs, numpy.array(line['y']) / duration_divider, "-", color=color, lw=lw, label=label)

        for line in vertical_lines:
            xs = [_ut_to_unloc_datetime(x) for x in line['x']]
            ax.plot(xs, numpy.array(line['y']) / duration_divider, ":", color=color)  # , lw=lw)

        assert (isinstance(ax, plt.Axes))

        if plot_journeys:
            xs = [_ut_to_unloc_datetime(x) for x in self.trip_departure_times]
            ys = self.trip_durations
            ax.plot(xs, numpy.array(ys) / duration_divider, "o", color="black", ms=8, label="journeys")
            if journey_letters is None:
                journey_letters = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"

                def cycle_journey_letters(journey_letters):
                    # cycle('ABCD') --> A B C D A B C D A B C D ...
                    saved = []
                    for element in journey_letters:
                        yield element
                        saved.append(element)
                    count = 1
                    while saved:
                        for element in saved:
                            yield element + str(count)
                        count += 1
                journey_letters_iterator = cycle_journey_letters(journey_letters)
            time_letters = {int(time): letter for letter, time in zip(journey_letters_iterator, self.trip_departure_times)}
            for x, y, letter in zip(xs, ys, journey_letters_iterator):
                walking = - self._walk_time_to_target / 30 if numpy.isfinite(self._walk_time_to_target) else 0
                ax.text(x + datetime.timedelta(seconds=(self.end_time_dep - self.start_time_dep) / 60),
                        (y + walking) / duration_divider, letter, va="top", ha="left")

        fill_between_x = []
        fill_between_y = []
        for line in slopes:
            xs = [_ut_to_unloc_datetime(x) for x in line['x']]
            fill_between_x.extend(xs)
            fill_between_y.extend(numpy.array(line["y"]) / duration_divider)

        ax.fill_between(fill_between_x, y1=fill_between_y, color=fill_color, alpha=alpha, label=label)

        ax.set_ylim(bottom=0)
        ax.set_ylim(ax.get_ylim()[0], ax.get_ylim()[1] * 1.05)

        if rcParams['text.usetex']:
            ax.set_xlabel(r"Departure time $t_{\mathrm{dep}}$")
        else:
            ax.set_xlabel("Departure time")

        ax.set_ylabel(r"Temporal distance $\tau$ (min)")
        if plot_journeys and return_letters:
            return ax, time_letters
        else:
            return ax

    def _temporal_distance_pdf(self):
        """
        Temporal distance probability density function.

        Returns
        -------
        non_delta_peak_split_points: numpy.array
        non_delta_peak_densities: numpy.array
            len(density) == len(temporal_distance_split_points_ordered) -1
        delta_peak_loc_to_probability_mass : dict
        """
        temporal_distance_split_points_ordered, norm_cdf = self.profile_block_analyzer._temporal_distance_cdf()
        delta_peak_loc_to_probability_mass = {}

        non_delta_peak_split_points = [temporal_distance_split_points_ordered[0]]
        non_delta_peak_densities = []
        for i in range(0, len(temporal_distance_split_points_ordered) - 1):
            left = temporal_distance_split_points_ordered[i]
            right = temporal_distance_split_points_ordered[i + 1]
            width = right - left
            prob_mass = norm_cdf[i + 1] - norm_cdf[i]
            if width == 0.0:
                delta_peak_loc_to_probability_mass[left] = prob_mass
            else:
                non_delta_peak_split_points.append(right)
                non_delta_peak_densities.append(prob_mass / float(width))
        assert (len(non_delta_peak_densities) == len(non_delta_peak_split_points) - 1)
        return numpy.array(non_delta_peak_split_points), \
               numpy.array(non_delta_peak_densities), \
               delta_peak_loc_to_probability_mass

    def get_temporal_distance_at(self, dep_time):
        return self.profile_block_analyzer.interpolate(dep_time)

    def get_temporal_distance_at(self, dep_time):
        return self.profile_block_analyzer


    @staticmethod
    def all_measures_and_names_as_lists():
        NPA = NodeProfileAnalyzerTime
        profile_summary_methods = [
            NPA.max_trip_duration,
            NPA.mean_trip_duration,
            NPA.median_trip_duration,
            NPA.min_trip_duration,
            NPA.max_temporal_distance,
            NPA.mean_temporal_distance,
            NPA.median_temporal_distance,
            NPA.min_temporal_distance,
            NPA.n_pareto_optimal_trips
        ]
        profile_observable_names = [
            "max_trip_duration",
            "mean_trip_duration",
            "median_trip_duration",
            "min_trip_duration",
            "max_temporal_distance",
            "mean_temporal_distance",
            "median_temporal_distance",
            "min_temporal_distance",
            "n_pareto_optimal_trips"
        ]
        assert (len(profile_summary_methods) == len(profile_observable_names))
        return profile_summary_methods, profile_observable_names



import io
import os
import unittest
import shutil

import networkx
import numpy
import pandas
import geojson

from gtfspy.gtfs import GTFS
from gtfspy import networks
from gtfspy.networks import ALL_STOP_TO_STOP_LINK_ATTRIBUTES
from gtfspy.route_types import BUS
from gtfspy import exports
from gtfspy.util import makedirs


class ExportsTest(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        """ This method is run once before executing any tests"""
        cls.gtfs_source_dir = os.path.join(os.path.dirname(__file__), "test_data")
        cls.G = GTFS.from_directory_as_inmemory_db(cls.gtfs_source_dir)

    # noinspection PyUnresolvedReferences
    def setUp(self):
        """This method is run once before _each_ test method is executed"""
        self.gtfs_source_dir = self.__class__.gtfs_source_dir
        self.gtfs = self.__class__.G
        self.extract_output_dir = os.path.join(self.gtfs_source_dir, "../", "test_gtfspy_extracts_8211231/")
        if not os.path.exists(self.extract_output_dir):
            makedirs(self.extract_output_dir)

    def tearDown(self):
        if os.path.exists(self.extract_output_dir):
            shutil.rmtree(self.extract_output_dir)

    def test_walk_network(self):
        walk_net = networks.walk_transfer_stop_to_stop_network(self.gtfs)
        self.assertGreater(len(walk_net.nodes()), 0)
        self.assertGreater(len(walk_net.edges()), 1)
        for form_node, to_node, data_dict in walk_net.edges(data=True):
            self.assertIn("d", data_dict)
            self.assertGreaterEqual(data_dict["d"], 0)
        threshold = 670
        walk_net = networks.walk_transfer_stop_to_stop_network(self.gtfs, max_link_distance=threshold)
        self.assertEqual(len(walk_net.edges()), 2)
        for form_node, to_node, data_dict in walk_net.edges(data=True):
            self.assertLess(data_dict['d'], threshold)

    def test_write_stop_to_stop_networks(self):
        exports.write_static_networks(self.gtfs, self.extract_output_dir)
        self.assertTrue(os.path.exists(self.extract_output_dir + "network_walk.edg"))
        self.assertTrue(os.path.exists(self.extract_output_dir + "network_bus.edg"))
        self.assertFalse(os.path.exists(self.extract_output_dir + "network_gondola.edg"))

    def test_write_combined_stop_to_stop_networks(self):
        output = os.path.join(self.extract_output_dir + "network_combined.edg")
        exports.write_combined_transit_stop_to_stop_network(self.gtfs, output)
        self.assertTrue(os.path.exists(output))

    def test_stop_to_stop_network_by_route_type(self):
        # test that distance works
        nxGraph = networks.stop_to_stop_network_for_route_type(self.gtfs,
                                                               BUS,
                                                               link_attributes=ALL_STOP_TO_STOP_LINK_ATTRIBUTES)
        self.assertTrue(isinstance(nxGraph, networkx.DiGraph), type(nxGraph))
        nodes = nxGraph.nodes(data=True)
        self.assertGreater(len(nodes), 0)
        for node in nodes:
            node_attributes = node[1]
            node_id = node[0]
            self.assertTrue(isinstance(node_id, (int, numpy.int_)))
            self.assertTrue("lat" in node_attributes)
            self.assertTrue("lon" in node_attributes)
            self.assertTrue("name" in node_attributes)
        edges = nxGraph.edges(data=True)
        self.assertGreater(len(edges), 0)

        at_least_one_shape_distance = False
        for from_I, to_I, linkData in edges:
            ds = linkData['distance_shape']
            self.assertTrue(isinstance(ds, int) or (ds is None),
                            "distance_shape should be either int or None (in case shapes are not available)")
            if isinstance(ds, int):
                at_least_one_shape_distance = True
            self.assertLessEqual(linkData['duration_min'], linkData["duration_avg"])
            self.assertLessEqual(linkData['duration_avg'], linkData["duration_max"])
            self.assertLessEqual(linkData['duration_median'], linkData["duration_max"])
            self.assertGreaterEqual(linkData['duration_median'], linkData["duration_min"])
            self.assertTrue(isinstance(linkData['d'], int), "straight line distance should always exist and be an int")
            self.assertGreaterEqual(linkData['d'], 0,
                               "straight line distance should be always greater than or equal to 0 (?)")
            n_veh = linkData["n_vehicles"]
            route_ids = linkData["route_I_counts"]
            route_ids_sum = sum([count for route_type, count in route_ids.items()])
            self.assertTrue(n_veh, route_ids_sum)

        self.assertTrue(at_least_one_shape_distance, "at least one shape distance should exist")

    def test_combined_stop_to_stop_transit_network(self):
        multi_di_graph = networks.combined_stop_to_stop_transit_network(self.gtfs)
        self.assertIsInstance(multi_di_graph, networkx.MultiDiGraph)
        for from_node, to_node, data in multi_di_graph.edges(data=True):
            self.assertIn("route_type", data)

    def test_temporal_network(self):
        temporal_pd = networks.temporal_network(self.gtfs)
        self.assertGreater(temporal_pd.shape[0], 10)

    def test_write_temporal_network(self):
        path = os.path.join(self.extract_output_dir, "combined.tnet")
        exports.write_temporal_network(self.gtfs, path, None, None)
        self.assertTrue(os.path.exists(path))
        df = pandas.read_csv(path)
        columns_should_exist = ["dep_time_ut", "arr_time_ut", "from_stop_I", "to_stop_I",
                                "route_type", "trip_I"]
        for col in columns_should_exist:
            self.assertIn(col, df.columns.values)

    def test_write_temporal_networks_by_route_type(self):
        exports.write_temporal_networks_by_route_type(self.gtfs, self.extract_output_dir)
        self.assertTrue(os.path.exists(os.path.join(self.extract_output_dir + "bus.tnet")))

    def test_write_gtfs_agencies(self):
        required_columns = 'agency_id,agency_name,agency_url,agency_timezone,agency_phone,agency_lang'.split(",")
        optional_columns = ['agency_lang', 'agency_phone', 'agency_fare_url', 'agency_email']
        self.__test_write_gtfs_table(exports._write_gtfs_agencies, required_columns, optional_columns)

    def test_write_gtfs_stops(self):
        required_columns = 'stop_id,stop_name,stop_desc,stop_lat,stop_lon'.split(",")
        optional_columns = ['stop_code', 'stop_desc', 'zone_id', 'stop_url', 'location_type', 'parent_station',
                               'stop_timezone', 'wheelchair_boarding']
        self.__test_write_gtfs_table(exports._write_gtfs_stops, required_columns, optional_columns)

    def test_write_gtfs_routes(self):
        required_columns = 'route_id,agency_id,route_short_name,route_long_name,route_desc,route_type'.split(",")
        optional_columns = ['route_desc', 'route_url', 'route_color', 'route_text_color']
        self.__test_write_gtfs_table(exports._write_gtfs_routes, required_columns, optional_columns)

    def test_write_gtfs_trips(self):
        required_columns = 'route_id,service_id,trip_id'.split(",")
        optional_columns = ['trip_headsign', 'trip_short_name', 'direction_id', 'block_id',
                                'shape_id', 'wheelchair_accessible', 'bikes_allowed']
        self.__test_write_gtfs_table(exports._write_gtfs_trips, required_columns, optional_columns)

    def test_write_gtfs_stop_times(self):
        required_columns = 'trip_id,arrival_time,departure_time,stop_id,stop_sequence'.split(",")
        optional_columns = ['stop_headsign', 'pickup_type', 'drop_off_type', 'shape_dist_traveled', 'timepoint']
        self.__test_write_gtfs_table(exports._write_gtfs_stop_times, required_columns, optional_columns)

    def test_write_gtfs_calendar(self):
        required_columns = 'service_id,monday,tuesday,wednesday,thursday,friday,saturday,sunday,' \
                            'start_date,end_date'.split(",")
        self.__test_write_gtfs_table(exports._write_gtfs_calendar, required_columns, [])
        in_memory_file = io.StringIO()
        exports._write_gtfs_calendar(self.gtfs, in_memory_file)
        in_memory_file.seek(0)
        df = pandas.read_csv(in_memory_file)
        self.assertTrue("-" not in str(df['start_date'][0]))
        self.assertTrue("-" not in str(df['end_date'][0]))
        self.assertTrue(len(str(df['start_date'][0])) == 8)
        self.assertTrue(len(str(df['start_date'][0])) == 8)

    def test_write_gtfs_calendar_dates(self):
        required_columns = 'service_id,date,exception_type'.split(",")
        self.__test_write_gtfs_table(exports._write_gtfs_calendar_dates, required_columns, [])

    def test_write_gtfs_shapes(self):
        required_columns = 'shape_id,shape_pt_lat,shape_pt_lon,shape_pt_sequence'.split(",")
        optional_columns = ['shape_dist_traveled']
        self.__test_write_gtfs_table(exports._write_gtfs_shapes, required_columns, optional_columns)

    def test_write_gtfs_transfers(self):
        required_columns = 'from_stop_id,to_stop_id,transfer_type'.split(",")
        optional_columns = ['min_transfer_time']
        self.__test_write_gtfs_table(exports._write_gtfs_transfers, required_columns, optional_columns)

    def test_write_gtfs_stop_distances(self):
        required_columns = 'from_stop_id,to_stop_id,d,d_walk'.split(",")
        optional_columns = []
        self.__test_write_gtfs_table(exports._write_gtfs_stop_distances, required_columns, optional_columns)

    def test_write_feed_info(self):
        required_columns = 'feed_publisher_name,feed_publisher_url,feed_lang'.split(",")
        columns_not_present = ['feed_start_date', 'feed_end_date', 'feed_version', 'feed_id']
        self.__test_write_gtfs_table(exports._write_gtfs_feed_info, required_columns, columns_not_present)

    def __test_write_gtfs_table(self, table_write_func, required_columns, optional_columns):
        """
        A helper method for testing writing of gtfs_table methods

        Parameters
        ----------
        table_write_func: function
        required_columns: list[str]
        optional_columns: list[str]
        """
        all_columns = set(required_columns)
        all_columns.update(set(optional_columns))

        in_memory_file = io.StringIO()
        table_write_func(self.gtfs, in_memory_file)
        in_memory_file.seek(0)
        header_columns = in_memory_file.readline().strip().split(",")
        for required_col in required_columns:
            self.assertIn(required_col, header_columns)
        for header_col in header_columns:
            self.assertIn(header_col, all_columns)

    def test_write_gtfs(self):
        # A simple import-output-import test"
        for ending in ["", ".zip"]:
            from gtfspy.import_gtfs import import_gtfs
            UUID = "36167f3012fe11e793ae92361f002671"
            sqlite_fname = "test_" + UUID + ".sqlite"
            test_output_dir = "./test_output_dir_" + UUID
            try:
                shutil.rmtree(test_output_dir)
            except FileNotFoundError:
                pass

            try:
                exports.write_gtfs(self.gtfs, test_output_dir + ending)
                self.assertTrue(os.path.exists(test_output_dir + ending))
                try:
                    G = import_gtfs(test_output_dir + ending, os.path.join(sqlite_fname))
                    self.assertTrue(os.path.exists(sqlite_fname))
                finally:
                    os.remove(sqlite_fname)
            finally:
                if ending == "":
                    shutil.rmtree(test_output_dir + ending)
                else:
                    os.remove(test_output_dir + ending)

    def test_write_stops_geojson(self):
        in_memory_file = io.StringIO()
        exports.write_stops_geojson(self.gtfs, in_memory_file)
        in_memory_file.seek(0)
        self.assertTrue(geojson.is_valid(in_memory_file.read(-1)))
        in_memory_file.seek(0)
        gjson = geojson.loads(in_memory_file.read(-1))
        gjson_properties = gjson['features'][0]['properties']
        self.assertIn("name", gjson_properties.keys())
        self.assertIn("stop_I", gjson_properties.keys())

    def test_write_sections_geojson(self):
        in_memory_file = io.StringIO()
        exports.write_sections_geojson(self.gtfs, in_memory_file)
        in_memory_file.seek(0)
        self.assertTrue(geojson.is_valid(in_memory_file.read(-1)))
        in_memory_file.seek(0)
        gjson = geojson.loads(in_memory_file.read(-1))
        gjson_properties = gjson['features'][0]['properties']
        self.assertIn("from_stop_I", gjson_properties.keys())
        self.assertIn("to_stop_I", gjson_properties.keys())
        self.assertIn("n_vehicles", gjson_properties.keys())
        self.assertIn("duration_avg", gjson_properties.keys())
        self.assertIn("route_I_counts", gjson_properties.keys())
        self.assertIn("route_type", gjson_properties.keys())

    def test_write_routes_geojson(self):
        in_memory_file = io.StringIO()
        exports.write_routes_geojson(self.gtfs, in_memory_file)
        in_memory_file.seek(0)
        self.assertTrue(geojson.is_valid(in_memory_file.read(-1)))
        in_memory_file.seek(0)
        gjson = geojson.loads(in_memory_file.read(-1))
        gjson_properties = gjson['features'][0]['properties']
        self.assertIn("route_type", gjson_properties.keys())
        self.assertIn("route_I", gjson_properties.keys())
        self.assertIn("route_name", gjson_properties.keys())



        # def test_clustered_stops_network(self):
    #     orig_net = networks.undirected_stop_to_stop_network_with_route_information(self.gtfs)
    #     aggregate_net = networks.aggregate_route_network(self.gtfs, 1000)
    #     self.assertGreater(len(orig_net.nodes()), len(aggregate_net.nodes()))
    #     self.assertTrue(isinstance(aggregate_net, networkx.Graph))
    #     for node in aggregate_net.nodes():
    #         for orig_node in node:
    #             self.assertIn(orig_node, orig_net.nodes())
    #     self.assertEquals(len(orig_net.nodes(), sum(map(lambda x: len(x), aggregate_net.nodes()))))
    #
    #     fake_aggregate_net = networks.aggregate_route_network(self.gtfs, 0)
    #     self.assertEquals(len(orig_net.nodes()), len(fake_aggregate_net.nodes()),
    #                       "no nodes should be using 0 distance")



import time

def timeit(method):
    """
    A Python decorator for printing out the execution time for a function.

    Adapted from:
    www.andreas-jung.com/contents/a-python-decorator-for-measuring-the-execution-time-of-methods
    """
    def timed(*args, **kw):
        time_start = time.time()
        result = method(*args, **kw)
        time_end = time.time()
        print('timeit: %r %2.2f sec (%r, %r) ' % (method.__name__, time_end-time_start, str(args)[:20], kw))
        return result

    return timed